class ISO6429:
    """
    Since 0.4.14
    @see https://en.wikipedia.org/wiki/ANSI_escape_code
    """

    # SGR (Select Graphic Rendition) Parameters

    SGR_RESET = 0
    SGR_BOLD = 1
    SGR_FAINT = 2  # 未广泛支持。
    SGR_ITALIC = 3  # 未广泛支持。有时视为反相显示。
    SGR_UNDERLINE = 4
    SGR_SLOW_BLINK = 5
    SGR_RAPID_BLINK = 6  # MS-DOS ANSI.SYS；每分钟150以上；未广泛支持。
    SGR_INVERT = 7
    SGR_HIDE = 8  # 未广泛支持。
    SGR_STRIKE = 9  # 未广泛支持。

    SGR_USE_DEFAULT_FONT = 10
    SGR_USE_FONT_1 = 11
    SGR_USE_FONT_2 = 12
    SGR_USE_FONT_3 = 13
    SGR_USE_FONT_4 = 14
    SGR_USE_FONT_5 = 15
    SGR_USE_FONT_6 = 16
    SGR_USE_FONT_7 = 17
    SGR_USE_FONT_8 = 18
    SGR_USE_FONT_9 = 19
    SGR_BLACKLETTER_FONT = 20  # 几乎无支持。

    SGR_DOUBLY_UNDERLINED = 21
    SGR_NOT_BOLD = 21

    SGR_NORMAL_INTENSITY = 22  # 正常颜色或强度
    SGR_NEITHER_ITALIC_NOR_BLACKLETTER = 23  # 非斜体、非尖角体
    SGR_NOT_UNDERLINED = 24  # 关闭下划线	去掉单双下划线。
    SGR_NOT_BLINKING = 25  # 关闭闪烁
    SGR_NOT_REVERSED = 27  # 关闭反显
    SGR_REVEAL = 28  # 关闭隐藏
    SGR_NOT_STRIKE = 29  # 关闭划除

    # 30–37	设置前景色	参见下面的颜色表。
    # 38	设置前景色	下一个参数是5;n或2;r;g;b，见下。
    # 39	默认前景色	由具体实现定义（按照标准）。

    SGR_SET_FOREGROUND_COLOR = 38

    SGR_FOREGROUND_COLOR_BLACK = 30
    SGR_FOREGROUND_COLOR_RED = 31
    SGR_FOREGROUND_COLOR_GREEN = 32
    SGR_FOREGROUND_COLOR_YELLOW = 33
    SGR_FOREGROUND_COLOR_BLUE = 34
    SGR_FOREGROUND_COLOR_MAGENTA = 35
    SGR_FOREGROUND_COLOR_CYAN = 36
    SGR_FOREGROUND_COLOR_WHITE = 37

    SGR_DEFAULT_FOREGROUND_COLOR = 39

    # 40–47	设置背景色	参见下面的颜色表。
    # 48	设置背景色	下一个参数是5;n或2;r;g;b，见下。
    # 49	默认背景色	由具体实现定义（按照标准）。

    SGR_Set_background_color = 48

    SGR_BACKGROUND_COLOR_BLACK = 40
    SGR_BACKGROUND_COLOR_RED = 41
    SGR_BACKGROUND_COLOR_GREEN = 42
    SGR_BACKGROUND_COLOR_YELLOW = 43
    SGR_BACKGROUND_COLOR_BLUE = 44
    SGR_BACKGROUND_COLOR_MAGENTA = 45
    SGR_BACKGROUND_COLOR_CYAN = 46
    SGR_BACKGROUND_COLOR_WHITE = 47

    SGR_DEFAULT_background_COLOR = 49

    # Implemented as "emoji variation selector" in mintty.
    # see https://github.com/mintty/mintty/wiki/Tips#text-attributes-and-rendering
    SGR_FRAMED = 51
    SGR_ENCIRCLED = 52
    SGR_NOT_FRAMED_OR_ENCIRCLED = 54

    SGR_OVERLINED = 53  # 上划线
    SGR_NOT_OVERLINED = 55  # 关闭上划线



from nehushtan.queue.NehushtanQueueTask import NehushtanQueueTask
from nehushtan.queue.NehushtanQueueTaskDelegate import NehushtanQueueTaskDelegate


class Test2NehushtanQueueTaskDelegate(NehushtanQueueTaskDelegate):

    def when_to_execute_task(self, task: NehushtanQueueTask, pid: int):
        self.logger.info(
            'Test2NehushtanQueueDelegate when_to_execute_task',
            {'task_id': task.get_task_reference(), 'pid': pid}
        )

    def when_task_executed(self, task: NehushtanQueueTask, pid: int):
        self.logger.info(
            'Test2NehushtanQueueDelegate when_task_executed',
            {'task_id': task.get_task_reference(), 'pid': pid}
        )

    def when_task_raised_exception(self, task: NehushtanQueueTask, exception: Exception):
        self.logger.exception(
            f'Test2NehushtanQueueDelegate when_task_raised_exception task_id={task.get_task_reference()}',
            exception
        )



import time

from nehushtan.helper.authenticate.TOTPKit import TOTPKit

if __name__ == '__main__':

    # shared_secret=b'BASE32SECRET2345AB=='

    # shared_secret = TOTPKit.generate_shared_secret_with_base32()
    # print(shared_secret)
    # print(shared_secret.decode())

    shared_secret = b'67YJGPVA63LAHZKYPJKT'
    print(shared_secret.lower())

    k = TOTPKit(shared_secret)

    last_p = ''
    last_count = 0
    for i in range(60):
        x = k.generate_current_totp_tuple()

        t = x[0]
        formatted_time = x[1]
        p = x[3]

        if last_p == p:
            last_count += 1
        else:
            print(f"last p {last_p} for {last_count} times")
            last_p = p
            last_count = 1

        print(f'applied on {x[0]}, available in [{x[1]}, {x[2]}): {x[3]}')

        time.sleep(1)

    print("end")



from psycopg2 import ProgrammingError

from nehushtan.logger.NehushtanLogger import NehushtanLogger
from nehushtan.postgres.PgSQLKit import PgSQLKit
from nehushtan.postgres.PgSQLKitConfig import PgSQLKitConfig
from nehushtan.postgres.PgSQLResultSet import PgSQLResultSet


class PgSQLTestEnv:
    def __init__(self):
        self.__logger = NehushtanLogger()

    def get_pg_kit_instance(self):
        return PgSQLKit(PgSQLKitConfig(
            {
                'host': '127.0.0.1',
                'port': 15432,
                'dbname': 'postgres',
                'user': 'postgres',
                'password': 'PGsql000',
                'autocommit': True,
            }
        ))

    def log(self, content, context=None):
        self.__logger.notice(content, context)

    def log_for_result_set(self, result_set: PgSQLResultSet):
        self.__logger.notice(
            f'QUERY: \n{result_set.get_query().decode()}\n',
            {
                "row_count": result_set.get_row_count(),
                "last_inserted_row_id": result_set.get_last_inserted_row_id(),
            }
        )
        try:
            if result_set.get_row_count() > 0:
                result_set.get_status_message()
                rows = result_set.fetch_all_remaining_rows()
                for row in rows:
                    self.__logger.info('> ', {"row": row})
        except ProgrammingError:
            pass



import re
from typing import Union, List

from nehushtan.cli.CSISequence import CSISequence
from nehushtan.cli.ISO6429 import ISO6429


class ANSITerminalHelper:
    """
    Since 0.4.13
    """

    def __init__(self):
        pass

    @staticmethod
    def write(text: Union[str, CSISequence], end: str = ''):
        # if type(text) is not str:
        #     print('<CSI-DEBUG>'+text.__str__()+'</CSI-DEBUG>')
        print(text.__str__(), end=end, flush=True)

    # 光标向指定的方向移动{\displaystyle n}n（默认1）格。如果光标已在屏幕边缘，则无效。

    def cursor_up(self, n: int = 1):
        self.write(CSISequence.csi_for_cursor_up(n))

    def cursor_down(self, n: int = 1):
        self.write(CSISequence.csi_for_cursor_down(n))

    def cursor_forward(self, n: int = 1):
        self.write(CSISequence.csi_for_cursor_forward(n))

    def cursor_back(self, n: int = 1):
        self.write(CSISequence.csi_for_cursor_back(n))

    def cursor_next_line(self, n: int = 1):
        """
        光标移动到下面第 n（默认1）行的开头。（非ANSI.SYS）
        """
        self.write(CSISequence.csi_for_cursor_next_line(n))

    def cursor_previous_line(self, n: int = 1):
        """
        光标移动到上面第 n（默认1）行的开头。（非ANSI.SYS）
        """
        self.write(CSISequence.csi_for_cursor_previous_line(n))

    def cursor_horizontal_absolute(self, n: int = 1):
        """
        光标移动到第 n（默认1）列。（非ANSI.SYS）
        """
        self.write(CSISequence.csi_for_cursor_horizontal_absolute(n))

    def cursor_move_to_position(self, row_number: int = 1, column_number: int = 1):
        """
        光标移动到第 row_number 行、第 column_number 列。
        值从1开始，且默认为1（左上角）。
        例如CSI ;5H和CSI 1;5H含义相同；
        CSI 17;H、CSI 17H和CSI 17;1H三者含义相同。
        """
        self.write(CSISequence.csi_for_cursor_move_to_position(row_number, column_number))

    def cursor_format_to_position(self, row_number: int = 1, column_number: int = 1):
        """
        Horizontal Vertical Position
        Same as CUP, but counts as a format effector function (like CR or LF)
            rather than an editor function (like CUD or CNL).
        This can lead to different handling in certain terminal modes.
        """
        self.write(CSISequence.csi_for_cursor_format_to_position(row_number, column_number))

    def erase_area(self, n: int = 0):
        """
        清除屏幕的部分区域。
        如果n是0（或缺失），则清除从光标位置到屏幕末尾的部分。
        如果n是1，则清除从光标位置到屏幕开头的部分。
        如果n是2，则清除整个屏幕（在DOS ANSI.SYS中，光标还会向左上方移动）。
        如果n是3，则清除整个屏幕，并删除回滚缓存区中的所有行（这个特性是xterm添加的，其他终端应用程序也支持）。
        """
        self.write(CSISequence.csi_for_erase_area(n))

    def erase_area_future(self):
        """
        清除从光标位置到屏幕末尾的部分
        """
        self.erase_area(0)

    def erase_area_past(self):
        """
        清除从光标位置到屏幕开头的部分
        """
        self.erase_area(1)

    def erase_area_all(self):
        """
        清除整个屏幕（在DOS ANSI.SYS中，光标还会向左上方移动）
        """
        self.erase_area(2)

    def erase_area_all_and_clean_buffer(self):
        """
        清除整个屏幕，并删除回滚缓存区中的所有行
        （这个特性是xterm添加的，其他终端应用程序也支持）
        """
        self.erase_area(3)

    def erase_line(self, n: int = 0):
        """
        清除行内的部分区域。
        如果n是0（或缺失），清除从光标位置到该行末尾的部分。
        如果n是1，清除从光标位置到该行开头的部分。
        如果n是2，清除整行。光标位置不变。
        """
        self.write(CSISequence.csi_for_erase_line(n))

    def erase_line_after(self):
        """
        清除从光标位置到该行末尾的部分
        """
        self.erase_line(0)

    def erase_line_before(self):
        """
        清除从光标位置到该行开头的部分
        """
        self.erase_line(1)

    def erase_line_all(self):
        """
        清除整行。光标位置不变
        """
        self.erase_line(2)

    def scroll_up(self, n: int = 1):
        """
        整页向上滚动n（默认1）行。新行添加到底部。（非ANSI.SYS）
        """
        self.write(CSISequence.csi_for_scroll_up(n))

    def scroll_down(self, n: int = 1):
        """
        整页向下滚动n（默认1）行。新行添加到顶部。（非ANSI.SYS）
        """
        self.write(CSISequence.csi_for_scroll_down(n))

    def select_graphic_rendition(self, sgr: Union[str, int, List[Union[str, int]]] = None):
        """
        选择图形再现（Select Graphic Rendition）
        """
        self.write(CSISequence.csi_for_select_graphic_rendition(sgr))

    def device_status_report(self):
        """
        Reports the cursor position (CPR) by transmitting `ESC[n;mR`, where n is the row and m is the column.)
        """
        self.write(CSISequence.csi_for_device_status_report())

    @staticmethod
    def parse_device_status_report():
        """
        read for the response of `device_status_report`
        """
        x = input()
        matches = re.match(r'\033\[(\d+);(\d+)R', x)
        if matches:
            return matches.group(1), matches.group(2)
        else:
            raise IOError(f'Expected `ESC[n;mR` but `{x}` read.')

    def print_colorful_text(self, text: str, foreground_color: int, background_color: int):
        """
        Since 0.4.14
        """
        self.select_graphic_rendition([foreground_color, background_color])
        self.write(text)
        self.select_graphic_rendition(ISO6429.SGR_RESET)



from typing import Literal, Optional

from nehushtan.paint.Paint import Paint


class ScoreDrawerOptions:
    def __init__(self,
                 default_ttf_font_file_path,
                 default_font_size: int
                 ):
        self.__default_ttf_font_file_path = default_ttf_font_file_path
        self.__default_font_size = default_font_size
        # self.__default_font = Paint.load_ttf_font(self.__default_ttf_font_file_path, self.__default_font_size)

        self.__title_font_file_path = None
        self.__title_font_size = None

        self.__title_left_component_font_file_path = None
        self.__title_left_component_font_size = None
        self.__title_middle_component_font_file_path = None
        self.__title_middle_component_font_size = None
        self.__title_right_component_font_file_path = None
        self.__title_right_component_font_size = None

        self.__score_font_path = None
        self.__score_font_size = None

        self.__lyric_font_path = None
        self.__lyric_font_size = None

        self.__lyric_prefix_font_path = None
        self.__lyric_prefix_font_size = None

        self.__total_cells_in_one_line = None

    def get_default_font_size(self):
        return self.__default_font_size

    def get_default_font(self):
        return Paint.load_ttf_font(self.__default_ttf_font_file_path, self.__default_font_size)

    def set_title_font_size(self, x):
        self.__title_font_size = x
        return self

    def set_title_font_file_path(self, x):
        self.__title_font_file_path = x
        return self

    def set_title_left_component_font_size(self, x):
        self.__title_left_component_font_size = x
        return self

    def set_title_left_component_font_file_path(self, x):
        self.__title_left_component_font_file_path = x
        return self

    def set_title_middle_component_font_size(self, x):
        self.__title_middle_component_font_size = x
        return self

    def set_title_middle_component_font_file_path(self, x):
        self.__title_middle_component_font_file_path = x
        return self

    def set_title_right_component_font_size(self, x):
        self.__title_right_component_font_size = x
        return self

    def set_title_right_component_font_file_path(self, x):
        self.__title_right_component_font_file_path = x
        return self

    def __get_font_for_title_x_component(self, x: Literal["left", "middle", "right"]):
        if x == "left":
            file = self.__title_left_component_font_file_path
            size = self.__title_left_component_font_size
        elif x == "right":
            file = self.__title_right_component_font_file_path
            size = self.__title_right_component_font_size
        else:
            file = self.__title_middle_component_font_file_path
            size = self.__title_middle_component_font_size

        if file is None:
            file = self.__title_font_file_path
        if file is None:
            file = self.__default_ttf_font_file_path

        if size is None:
            size = self.__title_font_size
        if size is None:
            size = self.__default_font_size

        return Paint.load_ttf_font(file, size)

    def get_font_for_title_left_component(self):
        return self.__get_font_for_title_x_component("left")

    def get_font_for_title_middle_component(self):
        return self.__get_font_for_title_x_component("middle")

    def get_font_for_title_right_component(self):
        return self.__get_font_for_title_x_component("right")

    def set_score_font_path(self, x):
        self.__score_font_path = x
        return self

    def set_score_font_size(self, x):
        self.__score_font_size = x
        return self

    def get_font_for_score(self):
        file = self.__score_font_path
        size = self.__score_font_size
        if file is None:
            file = self.__default_ttf_font_file_path
        if size is None:
            size = self.__default_font_size
        return Paint.load_ttf_font(file, size)

    def get_font_for_score_note(self):
        file = self.__score_font_path
        size = self.__score_font_size
        if file is None:
            file = self.__default_ttf_font_file_path
        if size is None:
            size = self.__default_font_size
        # size = int(size * 2 / 3.0)
        size = int(size * 0.5)
        return Paint.load_ttf_font(file, size)

    def set_lyric_font_path(self, x):
        self.__lyric_font_path = x
        return self

    def set_lyric_font_size(self, x):
        self.__lyric_font_size = x
        return self

    def set_lyric_prefix_font_path(self, x):
        self.__lyric_prefix_font_path = x
        return self

    def set_lyric_prefix_font_size(self, x):
        self.__lyric_prefix_font_size = x
        return self

    def get_font_for_lyric(self):
        file = self.__lyric_font_path
        size = self.__lyric_font_size
        if file is None:
            file = self.__default_ttf_font_file_path
        if size is None:
            size = self.__default_font_size
        return Paint.load_ttf_font(file, size)

    def get_font_for_lyric_prefix(self):
        file = self.__lyric_prefix_font_path
        size = self.__lyric_prefix_font_size
        if file is None:
            file = self.__lyric_font_path
        if size is None:
            size = self.__lyric_font_size
        if file is None:
            file = self.__default_ttf_font_file_path
        if size is None:
            size = self.__default_font_size
        return Paint.load_ttf_font(file, size)

    def set_total_cells_in_one_line(self, x: Optional[int]):
        self.__total_cells_in_one_line = x
        return self

    def get_total_cells_in_one_line(self) -> Optional[int]:
        return self.__total_cells_in_one_line



from abc import abstractmethod
from typing import Tuple, List, Sequence, Union, Dict


class NehushtanHTTPRoute:
    """
    Since 0.4.0
    """

    def __init__(self):
        self.matched_arguments = []
        self.matched_keyed_arguments = {}

    @abstractmethod
    def match_request(self, method: str, path: str) -> bool:
        pass

    @abstractmethod
    def get_filter_list(self) -> Sequence[Union[type, str]]:
        pass

    @abstractmethod
    def get_controller_target(self) -> Tuple[Union[type, str], str]:
        pass

    def get_matched_arguments(self) -> List[str]:
        return self.matched_arguments

    def get_matched_keyed_arguments(self) -> Dict[str, str]:
        return self.matched_keyed_arguments



from PySimpleGUI import LOOK_AND_FEEL_TABLE


class SimpleGuiTheme:

    def __init__(self, theme_name: str):
        self.__theme_name = theme_name

    def get_theme_name(self) -> str:
        return self.__theme_name

    @staticmethod
    def use_system_default():
        return SimpleGuiTheme("SystemDefault")

    @staticmethod
    def use_system_default_for_real():
        return SimpleGuiTheme("SystemDefaultForReal")

    @staticmethod
    def use_system_default_1():
        return SimpleGuiTheme("SystemDefault1")

    @staticmethod
    def use_material_1():
        return SimpleGuiTheme("Material1")

    @staticmethod
    def use_material_2():
        return SimpleGuiTheme("Material2")

    @staticmethod
    def use_reddit():
        return SimpleGuiTheme("Reddit")

    @staticmethod
    def use_topanga():
        return SimpleGuiTheme("Topanga")

    @staticmethod
    def use_green_tan():
        return SimpleGuiTheme("GreenTan")

    @staticmethod
    def use_dark():
        return SimpleGuiTheme("Dark")

    @staticmethod
    def use_light_green():
        return SimpleGuiTheme("LightGreen")

    @staticmethod
    def use_dark_2():
        return SimpleGuiTheme("Dark2")

    @staticmethod
    def use_black():
        return SimpleGuiTheme("Black")

    @staticmethod
    def use_tan():
        return SimpleGuiTheme("Tan")

    @staticmethod
    def use_tan_blue():
        return SimpleGuiTheme("TanBlue")

    @staticmethod
    def use_dark_tan_blue():
        return SimpleGuiTheme("DarkTanBlue")

    @staticmethod
    def use_dark_amber():
        return SimpleGuiTheme("DarkAmber")

    @staticmethod
    def use_dark_blue():
        return SimpleGuiTheme("DarkBlue")

    @staticmethod
    def use_reds():
        return SimpleGuiTheme("Reds")

    @staticmethod
    def use_green():
        return SimpleGuiTheme("Green")

    @staticmethod
    def use_blue_purple():
        return SimpleGuiTheme("BluePurple")

    @staticmethod
    def use_purple():
        return SimpleGuiTheme("Purple")

    @staticmethod
    def use_blue_mono():
        return SimpleGuiTheme("BlueMono")

    @staticmethod
    def use_green_mono():
        return SimpleGuiTheme("GreenMono")

    @staticmethod
    def use_brown_blue():
        return SimpleGuiTheme("BrownBlue")

    @staticmethod
    def use_bright_colors():
        return SimpleGuiTheme("BrightColors")

    @staticmethod
    def use_neutral_blue():
        return SimpleGuiTheme("NeutralBlue")

    @staticmethod
    def use_kayak():
        return SimpleGuiTheme("Kayak")

    @staticmethod
    def use_sandy_beach():
        return SimpleGuiTheme("SandyBeach")

    @staticmethod
    def use_teal_mono():
        return SimpleGuiTheme("TealMono")

    @staticmethod
    def use_default():
        return SimpleGuiTheme("Default")

    @staticmethod
    def use_default_1():
        return SimpleGuiTheme("Default1")

    @staticmethod
    def use_default_no_more_nagging():
        return SimpleGuiTheme("DefaultNoMoreNagging")

    @staticmethod
    def use_gray_gray_gray():
        return SimpleGuiTheme("GrayGrayGray")

    @staticmethod
    def use_light_blue():
        return SimpleGuiTheme("LightBlue")

    @staticmethod
    def use_light_grey():
        return SimpleGuiTheme("LightGrey")

    @staticmethod
    def use_light_grey_1():
        return SimpleGuiTheme("LightGrey1")

    @staticmethod
    def use_dark_brown():
        return SimpleGuiTheme("DarkBrown")

    @staticmethod
    def use_light_green_1():
        return SimpleGuiTheme("LightGreen1")

    @staticmethod
    def use_dark_grey():
        return SimpleGuiTheme("DarkGrey")

    @staticmethod
    def use_light_green_2():
        return SimpleGuiTheme("LightGreen2")

    @staticmethod
    def use_dark_grey_1():
        return SimpleGuiTheme("DarkGrey1")

    @staticmethod
    def use_dark_black():
        return SimpleGuiTheme("DarkBlack")

    @staticmethod
    def use_light_brown():
        return SimpleGuiTheme("LightBrown")

    @staticmethod
    def use_light_brown_1():
        return SimpleGuiTheme("LightBrown1")

    @staticmethod
    def use_dark_blue_1():
        return SimpleGuiTheme("DarkBlue1")

    @staticmethod
    def use_dark_brown_1():
        return SimpleGuiTheme("DarkBrown1")

    @staticmethod
    def use_dark_blue_2():
        return SimpleGuiTheme("DarkBlue2")

    @staticmethod
    def use_dark_brown_2():
        return SimpleGuiTheme("DarkBrown2")

    @staticmethod
    def use_dark_green():
        return SimpleGuiTheme("DarkGreen")

    @staticmethod
    def use_light_blue_1():
        return SimpleGuiTheme("LightBlue1")

    @staticmethod
    def use_light_purple():
        return SimpleGuiTheme("LightPurple")

    @staticmethod
    def use_light_blue_2():
        return SimpleGuiTheme("LightBlue2")

    @staticmethod
    def use_light_green_3():
        return SimpleGuiTheme("LightGreen3")

    @staticmethod
    def use_dark_blue_3():
        return SimpleGuiTheme("DarkBlue3")

    @staticmethod
    def use_light_green_4():
        return SimpleGuiTheme("LightGreen4")

    @staticmethod
    def use_light_green_5():
        return SimpleGuiTheme("LightGreen5")

    @staticmethod
    def use_light_brown_2():
        return SimpleGuiTheme("LightBrown2")

    @staticmethod
    def use_light_brown_3():
        return SimpleGuiTheme("LightBrown3")

    @staticmethod
    def use_light_blue_3():
        return SimpleGuiTheme("LightBlue3")

    @staticmethod
    def use_light_brown_4():
        return SimpleGuiTheme("LightBrown4")

    @staticmethod
    def use_dark_teal():
        return SimpleGuiTheme("DarkTeal")

    @staticmethod
    def use_dark_purple():
        return SimpleGuiTheme("DarkPurple")

    @staticmethod
    def use_light_green_6():
        return SimpleGuiTheme("LightGreen6")

    @staticmethod
    def use_dark_grey_2():
        return SimpleGuiTheme("DarkGrey2")

    @staticmethod
    def use_light_brown_6():
        return SimpleGuiTheme("LightBrown6")

    @staticmethod
    def use_dark_teal_1():
        return SimpleGuiTheme("DarkTeal1")

    @staticmethod
    def use_light_brown_7():
        return SimpleGuiTheme("LightBrown7")

    @staticmethod
    def use_dark_purple_1():
        return SimpleGuiTheme("DarkPurple1")

    @staticmethod
    def use_dark_grey_3():
        return SimpleGuiTheme("DarkGrey3")

    @staticmethod
    def use_light_brown_8():
        return SimpleGuiTheme("LightBrown8")

    @staticmethod
    def use_dark_blue_4():
        return SimpleGuiTheme("DarkBlue4")

    @staticmethod
    def use_light_blue_4():
        return SimpleGuiTheme("LightBlue4")

    @staticmethod
    def use_dark_teal_2():
        return SimpleGuiTheme("DarkTeal2")

    @staticmethod
    def use_dark_teal_3():
        return SimpleGuiTheme("DarkTeal3")

    @staticmethod
    def use_dark_purple_5():
        return SimpleGuiTheme("DarkPurple5")

    @staticmethod
    def use_dark_purple_2():
        return SimpleGuiTheme("DarkPurple2")

    @staticmethod
    def use_dark_blue_5():
        return SimpleGuiTheme("DarkBlue5")

    @staticmethod
    def use_light_grey_2():
        return SimpleGuiTheme("LightGrey2")

    @staticmethod
    def use_light_grey_3():
        return SimpleGuiTheme("LightGrey3")

    @staticmethod
    def use_dark_blue_6():
        return SimpleGuiTheme("DarkBlue6")

    @staticmethod
    def use_dark_blue_7():
        return SimpleGuiTheme("DarkBlue7")

    @staticmethod
    def use_light_brown_9():
        return SimpleGuiTheme("LightBrown9")

    @staticmethod
    def use_dark_purple_3():
        return SimpleGuiTheme("DarkPurple3")

    @staticmethod
    def use_light_brown_10():
        return SimpleGuiTheme("LightBrown10")

    @staticmethod
    def use_dark_purple_4():
        return SimpleGuiTheme("DarkPurple4")

    @staticmethod
    def use_light_blue_5():
        return SimpleGuiTheme("LightBlue5")

    @staticmethod
    def use_dark_teal_4():
        return SimpleGuiTheme("DarkTeal4")

    @staticmethod
    def use_light_teal():
        return SimpleGuiTheme("LightTeal")

    @staticmethod
    def use_dark_teal_5():
        return SimpleGuiTheme("DarkTeal5")

    @staticmethod
    def use_light_grey_4():
        return SimpleGuiTheme("LightGrey4")

    @staticmethod
    def use_light_green_7():
        return SimpleGuiTheme("LightGreen7")

    @staticmethod
    def use_light_grey_5():
        return SimpleGuiTheme("LightGrey5")

    @staticmethod
    def use_dark_brown_3():
        return SimpleGuiTheme("DarkBrown3")

    @staticmethod
    def use_light_brown1_1():
        return SimpleGuiTheme("LightBrown11")

    @staticmethod
    def use_dark_red():
        return SimpleGuiTheme("DarkRed")

    @staticmethod
    def use_dark_teal_6():
        return SimpleGuiTheme("DarkTeal6")

    @staticmethod
    def use_dark_brown_4():
        return SimpleGuiTheme("DarkBrown4")

    @staticmethod
    def use_light_yellow():
        return SimpleGuiTheme("LightYellow")

    @staticmethod
    def use_dark_green_1():
        return SimpleGuiTheme("DarkGreen1")

    @staticmethod
    def use_light_green_8():
        return SimpleGuiTheme("LightGreen8")

    @staticmethod
    def use_dark_teal_7():
        return SimpleGuiTheme("DarkTeal7")

    @staticmethod
    def use_dark_blue_8():
        return SimpleGuiTheme("DarkBlue8")

    @staticmethod
    def use_dark_blue_9():
        return SimpleGuiTheme("DarkBlue9")

    @staticmethod
    def use_dark_blue_10():
        return SimpleGuiTheme("DarkBlue10")

    @staticmethod
    def use_dark_blue_11():
        return SimpleGuiTheme("DarkBlue11")

    @staticmethod
    def use_dark_teal_8():
        return SimpleGuiTheme("DarkTeal8")

    @staticmethod
    def use_dark_red_1():
        return SimpleGuiTheme("DarkRed1")

    @staticmethod
    def use_light_brown_5():
        return SimpleGuiTheme("LightBrown5")

    @staticmethod
    def use_light_green_9():
        return SimpleGuiTheme("LightGreen9")

    @staticmethod
    def use_dark_green_2():
        return SimpleGuiTheme("DarkGreen2")

    @staticmethod
    def use_light_gray_1():
        return SimpleGuiTheme("LightGray1")

    @staticmethod
    def use_dark_grey_4():
        return SimpleGuiTheme("DarkGrey4")

    @staticmethod
    def use_dark_blue_12():
        return SimpleGuiTheme("DarkBlue12")

    @staticmethod
    def use_dark_purple_6():
        return SimpleGuiTheme("DarkPurple6")

    @staticmethod
    def use_dark_purple_7():
        return SimpleGuiTheme("DarkPurple7")

    @staticmethod
    def use_dark_blue_13():
        return SimpleGuiTheme("DarkBlue13")

    @staticmethod
    def use_dark_brown_5():
        return SimpleGuiTheme("DarkBrown5")

    @staticmethod
    def use_dark_green_3():
        return SimpleGuiTheme("DarkGreen3")

    @staticmethod
    def use_dark_black_1():
        return SimpleGuiTheme("DarkBlack1")

    @staticmethod
    def use_dark_grey_5():
        return SimpleGuiTheme("DarkGrey5")

    @staticmethod
    def use_light_brown_12():
        return SimpleGuiTheme("LightBrown12")

    @staticmethod
    def use_dark_teal_9():
        return SimpleGuiTheme("DarkTeal9")

    @staticmethod
    def use_dark_blue_14():
        return SimpleGuiTheme("DarkBlue14")

    @staticmethod
    def use_light_blue_6():
        return SimpleGuiTheme("LightBlue6")

    @staticmethod
    def use_dark_green_4():
        return SimpleGuiTheme("DarkGreen4")

    @staticmethod
    def use_dark_green_5():
        return SimpleGuiTheme("DarkGreen5")

    @staticmethod
    def use_dark_teal_10():
        return SimpleGuiTheme("DarkTeal10")

    @staticmethod
    def use_dark_grey_6():
        return SimpleGuiTheme("DarkGrey6")

    @staticmethod
    def use_dark_teal_11():
        return SimpleGuiTheme("DarkTeal11")

    @staticmethod
    def use_light_blue_7():
        return SimpleGuiTheme("LightBlue7")

    @staticmethod
    def use_light_green_10():
        return SimpleGuiTheme("LightGreen10")

    @staticmethod
    def use_dark_blue_15():
        return SimpleGuiTheme("DarkBlue15")

    @staticmethod
    def use_dark_blue_16():
        return SimpleGuiTheme("DarkBlue16")

    @staticmethod
    def use_dark_teal_12():
        return SimpleGuiTheme("DarkTeal12")

    @staticmethod
    def use_light_brown_13():
        return SimpleGuiTheme("LightBrown13")

    @staticmethod
    def use_dark_blue_17():
        return SimpleGuiTheme("DarkBlue17")

    @staticmethod
    def use_dark_brown_6():
        return SimpleGuiTheme("DarkBrown6")

    @staticmethod
    def use_dark_green_6():
        return SimpleGuiTheme("DarkGreen6")

    @staticmethod
    def use_dark_green_7():
        return SimpleGuiTheme("DarkGreen7")

    @staticmethod
    def use_dark_grey_7():
        return SimpleGuiTheme("DarkGrey7")

    @staticmethod
    def use_dark_red_2():
        return SimpleGuiTheme("DarkRed2")

    @staticmethod
    def use_light_grey_6():
        return SimpleGuiTheme("LightGrey6")

    @staticmethod
    def use_hot_dog_stand():
        return SimpleGuiTheme("HotDogStand")

    @staticmethod
    def use_dark_grey_8():
        return SimpleGuiTheme("DarkGrey8")

    @staticmethod
    def use_dark_grey_9():
        return SimpleGuiTheme("DarkGrey9")

    @staticmethod
    def use_dark_grey_10():
        return SimpleGuiTheme("DarkGrey10")

    @staticmethod
    def use_dark_grey_11():
        return SimpleGuiTheme("DarkGrey11")

    @staticmethod
    def use_dark_grey_12():
        return SimpleGuiTheme("DarkGrey12")

    @staticmethod
    def use_dark_grey_13():
        return SimpleGuiTheme("DarkGrey13")

    @staticmethod
    def use_dark_grey_14():
        return SimpleGuiTheme("DarkGrey14")

    @staticmethod
    def use_dark_grey_15():
        return SimpleGuiTheme("DarkGrey15")

    @staticmethod
    def use_dark_brown_7():
        return SimpleGuiTheme("DarkBrown7")

    @staticmethod
    def use_python():
        return SimpleGuiTheme("Python")

    @staticmethod
    def use_python_plus():
        return SimpleGuiTheme("PythonPlus")


if __name__ == '__main__':
    for key in LOOK_AND_FEEL_TABLE.keys():
        print(f'@staticmethod\ndef use{key}():\n    return SimpleGuiTheme("{key}")\n\n')



from nehushtan.score.NotationTextParser import NotationTextParser
from nehushtan.score.ScoreDrawer import ScoreDrawer
from nehushtan.score.ScoreDrawerOptions import ScoreDrawerOptions


def make(name: str):
    print("NAME: " + name)
    file = f"/Users/leqee/code/nehushtan/debug/score/book/{name}"
    parser = NotationTextParser(file=file)
    parser.parse()
    font1 = "/Users/leqee/code/Verbum/fonts/Arial Unicode.ttf"
    font2 = "/Users/leqee/code/Verbum/fonts/FangZhengKaiTi-GBK/FangZhengKaiTi-GBK-1.ttf"

    options = ScoreDrawerOptions(font1, 50)
    options.set_lyric_font_path(font1)
    options.set_lyric_font_size(40)
    options.set_score_font_size(40)
    options.set_title_left_component_font_size(35)
    options.set_title_right_component_font_size(35)
    options.set_title_middle_component_font_size(60)

    # options.set_total_cells_in_one_line(32)

    drawer = ScoreDrawer(parser.get_parsed_score_lines(), options)
    # drawer.show("test4")
    drawer.save(f"/Users/leqee/code/nehushtan/debug/score/png/{name}.png")


if __name__ == '__main__':
    for i in range(17):
        name = i + 1
        if name < 10:
            name = f'0{name}'
        else:
            name = f'{name}'

        make(name)



import random

from nehushtan.logger.NehushtanLogger import NehushtanLogger
from nehushtan.mysql.MySQLAnyTable import MySQLAnyTable
from nehushtan.mysql.MySQLKit import MySQLKit
from nehushtan.mysql.MySQLKitConfig import MySQLKitConfig
from tests.config import MYSQL_CONFIG

logger = NehushtanLogger('default')

config = MySQLKitConfig(MYSQL_CONFIG)
db = MySQLKit(config)

logger.info('prepare data set')

data_set = []
for i in range(100000):
    x1 = random.randint(100, 200)
    x2 = random.randint(400, 500)
    data_set.append([
        random.randint(1000, 2000),
        '2021-01-11',
        f'{x1}-{x2}',
        MySQLAnyTable.now()
    ])

logger.info('[raw] start with commit_immediately')

insert_sql = """
    INSERT IGNORE INTO sinri.dt_chat_spider_task(
        shop_id, rece_date, buyer_wang_id, ct_merits_create_time
    ) VALUES (%s, %s, %s, %s)
    """
last_row_id = db.raw_query_to_insert_many(insert_sql, data_set)

logger.info('[raw] end', {'last_row_id': last_row_id})

logger.info('prepare data set')

data_set = []
for i in range(100000):
    x1 = random.randint(100, 200)
    x2 = random.randint(400, 500)
    data_set.append([
        random.randint(1000, 2000),
        '2021-01-11',
        f'{x1}-{x2}',
        MySQLAnyTable.now()
    ])

logger.info('[model] start with commit_immediately')

afx = MySQLAnyTable(db, 'dt_chat_spider_task', 'sinri').insert_many_rows_with_matrix(
    ['shop_id', 'rece_date', 'buyer_wang_id', 'ct_merits_create_time'],
    data_set, with_ignore=True
).get_affected_rows()

logger.info('[model] end', {'afx': afx})

replace_sql = """
    REPLACE INTO sinri.dt_chat_spider_task(
        shop_id, rece_date, buyer_wang_id, ct_merits_create_time
    ) VALUES (%s, %s, %s, %s)
    """
last_row_id = db.raw_query_to_insert_many(replace_sql, data_set)

logger.info('--> [raw] end', {'last_row_id': last_row_id})

logger.info('fin')



import asyncio

from nehushtan.logger.NehushtanLogger import NehushtanLogger

logger = NehushtanLogger()


async def mirai(future_instance, p):
    logger.info(f"mirai[{p}] start")
    await asyncio.sleep(1)
    logger.info(f"mirai[{p}] slept for 1s")
    future_instance.set_exception(Exception("!"))
    # future_instance.set_result(p + " done")


def show_future_result(future_instance):
    if future_instance.cancelled():
        logger.warning("show_future_result::cancelled")
        return

    if not future_instance.done():
        logger.error("show_future_result::done not")
        return

    try:
        if future_instance.result():
            logger.notice("show_future_result::result", future_instance.result())
    except Exception as e:
        if future_instance.exception():
            logger.notice("show_future_result::exception", future_instance.exception().get_message())


async def main():
    loop = asyncio.get_running_loop()
    f = loop.create_future()

    f.add_done_callback(show_future_result)
    loop.create_task(mirai(f, "param"))

    logger.info("main start wait")
    logger.info(await f)

    if f.exception():
        logger.exception("exception", f.exception())


if __name__ == '__main__':
    asyncio.run(main())



from nehushtan.mysql.MySQLCondition import MySQLCondition

# Static Type Checking for Python, especially in PyCharm, is not so smart.
# If you mix a correct one into list, even others are wrong, the whole is right.

result = MySQLCondition.make_and(
    [
        2,
        MySQLCondition.make_equal('A', 1.5),
        MySQLCondition.make_in_array('B', ('1', '2')),
        1,
    ]
).organize_to_sql()
print(result)



from abc import ABC
from typing import List, Iterable

import pymysql

from nehushtan.mysql import constant
from nehushtan.mysql.MySQLCondition import MySQLCondition
from nehushtan.mysql.MySQLQueryResult import MySQLQueryResult
from nehushtan.mysql.MySQLViewMixin import MySQLViewMixin


class MySQLTableMixin(MySQLViewMixin, ABC):
    """
    This class is rewritten with the raw PyMySQL execute many method for multi-query.
    Since 0.1.17
    """

    @staticmethod
    def _build_on_duplicate_key_update_sql(on_duplicate_key_update_rows: dict):
        sql_parts = []
        for field, value in on_duplicate_key_update_rows.items():
            sql_parts.append(f'{field}={value}')
        sql_parts = ", ".join(sql_parts)
        return f'ON DUPLICATE KEY UPDATE {sql_parts}'

    def _modify_with_sql(self, sql_template: str, args: list = None, for_many_rows=False, commit_immediately=False):
        result = MySQLQueryResult(constant.MYSQL_QUERY_ROW_TYPE_UNKNOWN)
        cursor = None

        try:
            result.set_sql(sql_template)

            cursor = self.get_mysql_kit().get_raw_connection().cursor()

            if for_many_rows:
                cursor.executemany(sql_template, args)
            else:
                cursor.execute(sql_template, args)

            result.set_status(constant.MYSQL_QUERY_STATUS_EXECUTED)
            result.set_last_inserted_id(cursor.lastrowid)
            result.set_affected_rows(cursor.rowcount)
            if commit_immediately:
                self.get_mysql_kit().get_raw_connection().commit()

        except pymysql.MySQLError as e:
            result.set_status(constant.MYSQL_QUERY_STATUS_ERROR)
            result.set_error(f"MySQL Error {e.__class__}: [{e.args[0]}] {e.args[1]}")

        except Exception as pe:
            result.set_status(constant.MYSQL_QUERY_STATUS_ERROR)
            result.set_error(f"Python Error {pe.__class__}: {pe}")

        finally:
            if cursor is not None:
                cursor.close()
            return result

    def insert_one_row(
            self,
            row_dict: dict,
            commit_immediately: bool = False,
            on_duplicate_key_update_rows: dict = None,
            with_ignore: bool = False
    ):
        ignore_part = ''
        if with_ignore:
            ignore_part = ' IGNORE'

        table_expression = self.get_table_expression()

        field_names = []
        value_placeholders = []
        values = []
        for k, v in row_dict.items():
            field_names.append(f'`{k}`')
            value_placeholders.append("%s")
            values.append(v)

        field_names = ",".join(field_names)
        value_placeholders = ",".join(value_placeholders)

        on_duplicate_key_update_rows_part = ''
        if type(on_duplicate_key_update_rows) is dict:
            on_duplicate_key_update_rows_part = self._build_on_duplicate_key_update_sql(on_duplicate_key_update_rows)

        sql_template = f"""
            INSERT {ignore_part} INTO {table_expression} ({field_names}) 
            VALUES ({value_placeholders})
            {on_duplicate_key_update_rows_part}
        """

        return self._modify_with_sql(sql_template, values, False, commit_immediately)

    def replace_one_row(self, row_dict: dict, commit_immediately: bool = False):
        table_expression = self.get_table_expression()

        field_names = []
        value_placeholders = []
        values = []
        for k, v in row_dict.items():
            field_names.append(f'`{k}`')
            value_placeholders.append("%s")
            values.append(v)

        field_names = ",".join(field_names)
        value_placeholders = ",".join(value_placeholders)

        sql_template = f"""
            REPLACE INTO {table_expression} ({field_names})
            VALUES ({value_placeholders})
        """

        return self._modify_with_sql(sql_template, values, False, commit_immediately)

    def insert_many_rows_with_dicts(
            self,
            row_dicts: List[dict],
            commit_immediately: bool = False,
            on_duplicate_key_update_rows: dict = None,
            with_ignore: bool = False
    ):
        ignore_part = ''
        if with_ignore:
            ignore_part = ' IGNORE'

        table_expression = self.get_table_expression()

        field_names = []
        placeholders = []
        matrix = []

        first_row = row_dicts[0]
        for k, v in first_row.items():
            field_names.append(f'`{k}`')
            placeholders.append("%s")

        for row_dict in row_dicts:
            x = []
            for k, v in row_dict.items():
                x.append(v)
            matrix.append(x)

        field_names = ",".join(field_names)
        placeholders = ",".join(placeholders)

        on_duplicate_key_update_rows_part = ''
        if type(on_duplicate_key_update_rows) is dict:
            on_duplicate_key_update_rows_part = self._build_on_duplicate_key_update_sql(on_duplicate_key_update_rows)

        sql_template = f"""
            INSERT {ignore_part} INTO {table_expression} ({field_names})
            VALUES ({placeholders})
            {on_duplicate_key_update_rows_part}
        """

        return self._modify_with_sql(sql_template, matrix, True, commit_immediately)

    def replace_many_rows_with_dicts(self, row_dicts: List[dict], commit_immediately: bool = False):
        table_expression = self.get_table_expression()

        field_names = []
        placeholders = []
        matrix = []

        first_row = row_dicts[0]
        for k, v in first_row.items():
            field_names.append(f'`{k}`')
            placeholders.append("%s")

        for row_dict in row_dicts:
            x = []
            for k, v in row_dict.items():
                x.append(v)
            matrix.append(x)

        field_names = ",".join(field_names)
        placeholders = ",".join(placeholders)

        sql_template = f"""
            REPLACE INTO {table_expression} ({field_names})
            VALUES ({placeholders})
        """

        return self._modify_with_sql(sql_template, matrix, True, commit_immediately)

    def insert_many_rows_with_matrix(
            self,
            fields: List[str],
            row_matrix: List[List],
            commit_immediately: bool = False,
            on_duplicate_key_update_rows: dict = None,
            with_ignore: bool = False
    ):
        ignore_part = ''
        if with_ignore:
            ignore_part = ' IGNORE'

        table_expression = self.get_table_expression()

        field_names = ",".join([f'`{k}`' for k in fields])
        placeholders = ["%s"] * len(fields)
        placeholders = ",".join(placeholders)

        on_duplicate_key_update_rows_part = ''
        if type(on_duplicate_key_update_rows) is dict:
            on_duplicate_key_update_rows_part = self._build_on_duplicate_key_update_sql(on_duplicate_key_update_rows)

        sql_template = f"""
            INSERT {ignore_part} INTO {table_expression} ({field_names})
            VALUES ({placeholders})
            {on_duplicate_key_update_rows_part}
        """

        return self._modify_with_sql(sql_template, row_matrix, True, commit_immediately)

    def replace_many_rows_with_matrix(
            self,
            fields: List[str],
            row_matrix: List[List],
            commit_immediately: bool = False
    ):
        table_expression = self.get_table_expression()

        field_names = ",".join([f'`{k}`' for k in fields])
        placeholders = ["%s"] * len(fields)
        placeholders = ",".join(placeholders)

        sql_template = f"""
            REPLACE INTO {table_expression} ({field_names})
            VALUES ({placeholders})
        """

        return self._modify_with_sql(sql_template, row_matrix, True, commit_immediately)

    def write_rows_with_raw_selection_sql(
            self,
            write_type: str,
            fields: List[str],
            selection_sql: str,
            with_ignore: bool = False,
            on_duplicate_key_update_rows: dict = None,
            commit_immediately: bool = False
    ):
        """
        < Experimental Method, Since 0.1.10>
        It is a low level method for
        [INSERT ... SELECT Statement](https://dev.mysql.com/doc/refman/8.0/en/insert-select.html)
        and [REPLACE Statement](https://dev.mysql.com/doc/refman/8.0/en/replace.html)
        Be Careful when use this.
        :param write_type: INSERT or REPLACE
        :param fields:
        :param selection_sql: 'SELECT ...' If select from one table, you may use MySQLTableSelection::generate_sql()
        :param with_ignore:
        :param on_duplicate_key_update_rows:
        :param commit_immediately:
        :return:
        """

        ignore_part = ''
        if with_ignore:
            ignore_part = ' IGNORE'

        fields_sql = ",".join([f'`{k}`' for k in fields])

        on_duplicate_key_update_rows_part = ''
        if type(on_duplicate_key_update_rows) is dict:
            on_duplicate_key_update_rows_part = self._build_on_duplicate_key_update_sql(on_duplicate_key_update_rows)

        sql_template = f"""
            {write_type}{ignore_part} INTO {self.get_table_expression()} ({fields_sql}) 
            {selection_sql}
            {on_duplicate_key_update_rows_part}
        """

        return self._modify_with_sql(sql_template, commit_immediately=commit_immediately)

    def update_rows(
            self,
            conditions: Iterable[MySQLCondition],
            modifications: dict,
            commit_immediately: bool = False,
            with_ignore: bool = False,
            sort_expression: str = '',
            limit: int = 0
    ):
        """
        [UPDATE Statement](https://dev.mysql.com/doc/refman/8.0/en/update.html)
        :param conditions:
        :param modifications:
        :param commit_immediately:
        :param with_ignore:
        :param sort_expression:
        :param limit:
        :return:
        """
        condition_sql = MySQLCondition.build_sql_component(conditions)

        modify_pairs = []
        for k, v in modifications.items():
            modify_pairs.append(f'`{k}`=' + self.get_mysql_kit().quote(v))
        modify_pairs = ",".join(modify_pairs)

        ignore = ''
        if with_ignore:
            ignore = ' IGNORE'

        sql = f"UPDATE{ignore} {self.get_table_expression()} SET {modify_pairs} WHERE {condition_sql}"

        if len(sort_expression) > 0:
            sql = f"{sql} ORDER BY {sort_expression}"

        if limit > 0:
            sql = f"{sql} LIMIT {limit}"

        return self._modify_with_sql(sql, None, False, commit_immediately)

    def delete_rows(
            self,
            conditions: Iterable[MySQLCondition],
            commit_immediately: bool = False,
            with_ignore: bool = False,
            sort_expression: str = '',
            limit: int = 0
    ):
        """
        [DELETE Statement](https://dev.mysql.com/doc/refman/8.0/en/delete.html)
        :param conditions:
        :param commit_immediately:
        :param with_ignore:
        :param sort_expression:
        :param limit:
        :return:
        """

        condition_sql = MySQLCondition.build_sql_component(conditions)

        ignore = ''
        if with_ignore:
            ignore = ' IGNORE'

        sql = f"DELETE{ignore} FROM {self.get_table_expression()} WHERE {condition_sql}"

        if len(sort_expression) > 0:
            sql = f"{sql} ORDER BY {sort_expression}"

        if limit > 0:
            sql = f"{sql} LIMIT {limit}"

        return self._modify_with_sql(sql, None, False, commit_immediately)

    def truncate(self):
        sql_template = f"TRUNCATE {self.get_table_expression()}"
        return self._modify_with_sql(sql_template)



import time
from random import random
from typing import List

from nehushtan.logger.NehushtanLogger import NehushtanLogger
from nehushtan.queue.NehushtanQueueDelegate import NehushtanQueueDelegate
from nehushtan.queue.NehushtanQueueTask import NehushtanQueueTask
from nehushtan.queue.situation.NoNextTaskSituation import NoNextTaskSituation
from tests.queue.plan2.Test2NehushtanQueueTask import Test2NehushtanQueueTask


class Test2NehushtanQueueDelegate(NehushtanQueueDelegate):

    def should_kill_any_worker_processes(self) -> list:
        return []

    def when_killed_worker_process(self, task_reference, not_found=False, worker_pid: int = None):
        self.logger.notice(
            'when_killed_worker_process',
            {'task_reference': task_reference, 'not_found': not_found, 'worker_pid': worker_pid}
        )

    def handle_command_queue(self) -> int:
        self.logger.info('handle_command_queue, actually do nothing')
        return 0

    def __init__(self, config_dictionary: dict = None, logger: NehushtanLogger = None):
        super().__init__(config_dictionary, logger)
        self.last_task_id = 1000

    def when_loop_reports_error(self, error_message: str, cause_exception: Exception = None):
        if cause_exception:
            self.logger.exception(error_message, cause_exception)
        else:
            self.logger.error(f'Test2NehushtanQueueDelegate when_loop_reports_error: {error_message}')

    def when_loop_terminates(self):
        self.logger.critical('Test2NehushtanQueueDelegate when_loop_terminates')

    def when_loop_should_not_run(self):
        self.logger.warning('Test2NehushtanQueueDelegate when_loop_should_not_run sleep 10')
        time.sleep(10)

    def check_next_task_candidates(self) -> List[NehushtanQueueTask]:
        self.logger.info('Test2NehushtanQueueDelegate check_next_task')
        if random() > 0.7:
            self.last_task_id += 1
            self.logger.info('Test2NehushtanQueueDelegate check_next_task -> create new task')
            return [Test2NehushtanQueueTask(task_id=self.last_task_id)]
        else:
            raise NoNextTaskSituation('No task to execute')

    def when_no_task_to_do(self):
        self.logger.info('Test2NehushtanQueueDelegate when_no_task_to_do sleep 5')
        time.sleep(5)

    def when_task_not_executable(self, task: NehushtanQueueTask):
        self.logger.info('Test2NehushtanQueueDelegate when_task_not_executable', {'task_id': task.get_task_reference()})



import imaplib
import re
import warnings
from typing import Iterable, List

from nehushtan.logger.NehushtanLogger import NehushtanLogger, NehushtanLogLevel
from nehushtan.mail.rfc3501.SearchCommandKit import SearchCommandKit
from nehushtan.mail.rfc822.NehushtanEmailMessage import NehushtanEmailMessage


class IMAPAgent:
    """
    Greatly Changed between 0.4.7 and 0.4.8
    """

    FETCH_METHOD_RFC822 = '(RFC822)'
    FETCH_METHOD_RFC822_HEADER = '(RFC822.HEADER)'
    FETCH_METHOD_RFC822_WITH_UID = '(UID RFC822)'
    FETCH_METHOD_RFC822_HEADER_WITH_UID = '(UID RFC822.HEADER)'

    """
    Since 0.1.13
    [Experimental, Not Fully Completed]
    """
    STATUS_NAME_MESSAGES = 'MESSAGES'  # 邮箱中的邮件数。
    STATUS_NAME_RECENT = 'RECENT'  # 设置了 .ecent 标志的消息数。
    STATUS_NAME_UIDNEXT = 'UIDNEXT'  # 邮箱的下一个唯一标识符值。
    STATUS_NAME_UIDVALIDITY = 'UIDVALIDITY'  # 邮箱的唯一标识符有效性值。
    STATUS_NAME_UNSEEN = 'UNSEEN'  # 没有设置 .een 标志的消息数。

    def __init__(self, host: str, port: int, use_ssl: bool, logger: NehushtanLogger = None):
        if use_ssl:
            self._connection = imaplib.IMAP4_SSL(host, port)
        else:
            self._connection = imaplib.IMAP4(host, port)

        if logger:
            self.logger = logger
        else:
            self.logger = NehushtanLogger(log_level=NehushtanLogLevel.CRITICAL,)

    def login(self, username: str, password: str):
        self._connection.login(username, password)
        return self

    def logout(self):
        self._connection.logout()

    def list_mail_boxes(self, directory='""', pattern='*'):
        """
        检索帐户可用的邮箱
        :param directory:
        :param pattern:
        :return: 一个字符串序列，包含每个邮箱的 `标志`，`层次结构分隔符` 和 `邮箱名称`
        """
        response_code, boxes = self._connection.list(directory, pattern)
        if response_code != 'OK':
            raise Exception(f'IMAPAgent cannot fetch box list. Code = {response_code} Data = {boxes}')
        return boxes

    @staticmethod
    def parse_box_string_to_tuple(box_string):
        """

        :param box_string:
        :return: A tuple with parsed components (flags, delimiter, mailbox_name)
        """
        list_response_pattern = re.compile(
            r'.(?P<flags>.*?). "(?P<delimiter>.*)" (?P<name>.*)'
        )

        match = list_response_pattern.match(box_string.decode('utf-8'))
        flags, delimiter, mailbox_name = match.groups()
        mailbox_name = mailbox_name.strip('"')
        return flags, delimiter, mailbox_name

    def show_status(self, box: str, status_name_array: Iterable[str]):
        status_name_array_str = '(' + " ".join(status_name_array) + ')'
        response_code, x = self._connection.status(f'"{box}"', status_name_array_str)
        if response_code != 'OK':
            raise Exception(f"IMAPAgent cannot fetch box status info for Box {box} {status_name_array_str}")

        # print(x)
        status_response_pattern = re.compile(
            r'"?(?P<box_name>.*)"? \((?P<status_string>.*)\)'
        )
        dd = {}
        for row in x:
            # print(f'row to re: {row}')
            match = status_response_pattern.match(row.decode('utf-8'))
            # print('match is ', match)
            box_name, status_string = match.groups()
            parts = status_string.split(' ')
            d = {}
            key = None
            for part in parts:
                if key is None:
                    key = part
                else:
                    d[key] = int(part)
                    key = None
            dd[box_name] = d
        return dd

    def select_mailbox(self, box: str, readonly: bool = False):
        """

        :param box:
        :param readonly:
        :return: The Number of Total Mails
        """
        response_code, data = self._connection.select(box, readonly)
        if response_code != 'OK':
            raise Exception(f"IMAPAgent select_mailbox {box} failed: {response_code} with Data: {data}")
        return int(data[0])

    def search_mails_for_message_id(self, criteria, charset=None):
        """
        :param criteria:
        :param charset:
        :return: Message ID array
        """
        warnings.warn('NOT FRIENDLY WITH NON-ASCII')
        response_code, data = self._connection.search(charset, criteria)
        if response_code != 'OK':
            raise Exception(f"IMAPAgent search_mails_in_mailbox failed: {response_code} with Data: {data}")
        message_id_array = data[0].decode('utf-8').split(' ')
        return message_id_array

    def search_mail_for_uid(self, search: SearchCommandKit):
        command, arguments, literal = search.build()
        if literal is not None:
            self._connection.literal = literal
        response_code, data = self._connection.uid(command, *arguments)
        if response_code != 'OK':
            raise Exception(f"IMAPAgent search_mails_in_mailbox failed: {response_code} with Data: {data}")
        uid_array: List[str] = data[0].decode().split(' ')
        # uid_array = data[0].decode(search.get_charset()).split(' ')

        if len(uid_array) == 1 and uid_array[0] == '':
            return []
        return uid_array

    def fetch_mail_with_uid(self, uid: str, message_parts: str):
        arguments = [uid, message_parts]
        response_code, data = self._connection.uid('FETCH', *arguments)
        if response_code != 'OK':
            raise Exception(f"IMAPAgent fetch_mail_using_uid failed: {response_code} with Data: {data}")
        return data

    def fetch_mail_with_message_id(self, message_id: str, message_parts: str) -> list:
        """

        :param message_id:
        :param message_parts:  such as '(BODY.PEEK[HEADER] FLAGS)' for subject
        :return:
        """
        response_code, data = self._connection.fetch(message_id, message_parts)
        if response_code != 'OK':
            raise Exception(f"IMAPAgent fetch_mail failed: {response_code} with Data: {data}")
        return data

    # # 字符编码转换
    # @staticmethod
    # def decode_str(str_in):
    #     value, charset = decode_header(str_in)[0]
    #     if charset:
    #         value = value.decode(charset)
    #     return value

    def fetch_mail_with_message_id_as_nem(
            self,
            message_id: str,
            headers_only: bool = False
    ) -> NehushtanEmailMessage:
        """
        Since 0.4.8
        """
        fetch_method = self.FETCH_METHOD_RFC822
        if headers_only:
            fetch_method = self.FETCH_METHOD_RFC822_HEADER

        rfc822 = self.fetch_mail_with_message_id(message_id, fetch_method)
        raw_mail_text: bytes = rfc822[0][1]
        return NehushtanEmailMessage.parse_bytes(raw_mail_text)

    def fetch_mail_with_uid_as_nem(
            self,
            uid: str,
            headers_only: bool = False
    ) -> NehushtanEmailMessage:
        """
        Since 0.4.8
        """
        fetch_method = self.FETCH_METHOD_RFC822_WITH_UID
        if headers_only:
            fetch_method = self.FETCH_METHOD_RFC822_HEADER_WITH_UID

        rfc822 = self.fetch_mail_with_uid(uid, fetch_method)
        raw_mail_text: bytes = rfc822[0][1]
        return NehushtanEmailMessage.parse_bytes(raw_mail_text)



from PIL import ImageFont

from nehushtan.paint.Paint import Paint

if __name__ == '__main__':
    # use a truetype font
    font = ImageFont.truetype("/Users/leqee/code/Verbum/fonts/OradanoGSRR.ttf", 50)

    paint = Paint(400, 600, background_color="white")

    paint.rectangle(
        (
            (10, 10),
            (391, 591)
        ),
        outline="red"
    )

    paint.text(
        (50, 50, 200, 100),
        fill="red",
        font=font,
        align="center",
        text="なりすまし"
    )

    paint.show("test2")



import json
import re
from typing import Literal, Optional


class ScoreUnit:
    # CHAR_ACCIDENTAL_SHARP = "#"
    # CHAR_ACCIDENTAL_FLAT = "b"  # ♭
    # CHAR_ACCIDENTAL_NATURAL = "n"  # ♮
    # CHAR_DOT = "."
    # CHAR_TRIPLET = "%"
    # CHAR_ABOVE_POINT = ">"
    # CHAR_BELOW_POINT = "<"
    # CHAR_UNDER_LINE = "_"
    # CHAR_RIGHT_LINE = "-"
    # CHAR_TIE_OR_SLUR_BEGIN = "("
    # CHAR_TIE_OR_SLUR_END = ")"
    # CHAR_FERMATA = "~"
    #
    # CHAR_REMARK_AS_LEFT = ":>"
    # CHAR_REMARK_AS_RIGHT = ":<"
    # CHAR_REMARK_AS_CENTER = ":="

    def __init__(self,
                 note: Literal[
                     "0", "1", "2", "3", "4", "5", "6", "7", "|", "||", "||:", ":||", "|:", ":|", "(", ")", ")3"],
                 under_line_count: int = 0,
                 right_line_count: int = 0,
                 above_point_count: int = 0,
                 below_point_count: int = 0,
                 dot: bool = False,
                 accidental: Optional[Literal["#", "b", "n"]] = None,
                 # ties_or_slurs_begins: bool = False,
                 # ties_or_slurs_ends: bool = False,
                 fermata: bool = False,
                 # triplet: bool = False,
                 remark_style: Optional[Literal[":>", ":=", ":<"]] = None,
                 remark_text: Optional[str] = None
                 ):
        self.__note = note
        self.__under_line_count = under_line_count
        self.__right_line_count = right_line_count
        self.__above_point_count = above_point_count
        self.__below_point_count = below_point_count
        self.__dot = dot
        self.__accidental = accidental
        # self.__ties_or_slurs_begins = ties_or_slurs_begins
        # self.__ties_or_slurs_ends = ties_or_slurs_ends
        self.__fermata = fermata
        # self.__triplet = triplet
        self.__remark_style = remark_style
        self.__remark_text = remark_text

    def __repr__(self):
        return {
            "note": self.__note,
            "under_line_count": self.__under_line_count,
            "right_line_count": self.__right_line_count,
            "above_point_count": self.__above_point_count,
            "below_point_count": self.__below_point_count,
            "dot": self.__dot,
            "accidental": self.__accidental,
            "fermata": self.__fermata,
            # "triplet": self.__triplet,
            "remark": {
                "style": self.__remark_style,
                "text": self.__remark_text
            }
        }

    def get_accidental(self):
        return self.__accidental

    def get_note(self):
        return self.__note

    def get_under_line_count(self):
        return self.__under_line_count

    def get_right_line_count(self):
        return self.__right_line_count

    def get_above_point_count(self):
        return self.__above_point_count

    def get_below_point_count(self):
        return self.__below_point_count

    def get_dot(self):
        return self.__dot

    def get_fermata(self):
        return self.__fermata

    # def get_triplet(self):
    #     return self.__triplet

    def get_remark_style(self):
        return self.__remark_style

    def get_remark_text(self):
        return self.__remark_text

    def __str__(self):
        x = self.__repr__()
        return json.dumps(x)

    def get_cell_needed(self):
        if self.__note in ("(", ")"):
            return 0
        else:
            x = 1
            if self.__right_line_count > 0:
                x += self.__right_line_count
            if self.__dot:
                x += 1 * 0
            return x

    @staticmethod
    def parse_code_to_score(the_code: str):
        if the_code is None or len(the_code.strip()) == 0:
            raise SyntaxError("score code error: empty")

        code = the_code

        remark_style = None
        remark_text = None

        r0 = re.compile("^(.+)((:[<>=])(.+))$").match(code)
        if r0:
            code = r0.group(1)
            remark_style = r0.group(3)
            remark_text = r0.group(4)

        if code in ("||:", "|:", "|", ":|", ":||", "||", "(", ")", ")3"):
            return ScoreUnit(note=code, remark_style=remark_style, remark_text=remark_text)

        r1 = re.compile("^([b#n]?)([0-7])(.*)$").match(code)
        if not r1:
            raise SystemError("score note error: " + code)

        # print(f"debug: 0: {r1.group(0)} 1: {r1.group(1)} 2: {r1.group(2)} 3: {r1.group(3)}")

        accidental = r1.group(1)
        if accidental == "":
            accidental = None
        note = r1.group(2)
        code = r1.group(3)

        if code is None or len(code) == 0:
            # print("biaozhun si fen yin fu")
            return ScoreUnit(note=note, accidental=accidental, remark_style=remark_style, remark_text=remark_text)

        above_point_count = 0
        below_point_count = 0
        under_line_count = 0
        right_line_count = 0
        dot = False
        fermata = False
        # triplet = False

        r2 = re.compile(">+").search(code)
        if r2:
            above_point_count = len(r2.group(0))
        r2 = re.compile("<+").search(code)
        if r2:
            below_point_count = len(r2.group(0))
        r2 = re.compile("_+").search(code)
        if r2:
            under_line_count = len(r2.group(0))
        r2 = re.compile("-+").search(code)
        if r2:
            right_line_count = len(r2.group(0))

        r2 = re.compile("\.").search(code)
        if r2:
            # print(f"debug dot is {dot} for {code}")
            dot = True
        r2 = re.compile("~").search(code)
        if r2:
            fermata = True
        # r2 = re.compile("%").search(code)
        # if r2:
        #     triplet = True

        return ScoreUnit(
            note=note, accidental=accidental, remark_style=remark_style, remark_text=remark_text,
            above_point_count=above_point_count,
            below_point_count=below_point_count,
            under_line_count=under_line_count,
            right_line_count=right_line_count,
            dot=dot,
            fermata=fermata
            # triplet=triplet
        )



from nehushtan.mail.IMAPAgent import IMAPAgent
from tests.config import IMAP_CONFIG

status_name_array = (
    IMAPAgent.STATUS_NAME_MESSAGES,
    IMAPAgent.STATUS_NAME_RECENT,
    IMAPAgent.STATUS_NAME_UIDNEXT,
    IMAPAgent.STATUS_NAME_UIDVALIDITY,
    IMAPAgent.STATUS_NAME_UNSEEN,
)

imap_agent = IMAPAgent(IMAP_CONFIG['imap_server'], IMAP_CONFIG['imap_port'], True)
imap_agent.login(IMAP_CONFIG['username'], IMAP_CONFIG['password'])
boxes = imap_agent.list_mail_boxes()
for box in boxes:
    box_tuple = IMAPAgent.parse_box_string_to_tuple(box)
    # print(box,box_tuple)

    box_name = box_tuple[2]

    status = imap_agent.show_status(box_name, status_name_array)
    print('Box Status: ', status)

data = imap_agent.select_mailbox('INBOX')
print(f'Total Mails: {data}')

message_id_array = imap_agent.search_mails_for_message_id('ALL')
print(message_id_array)

for message_id in message_id_array:
    print(f"MESSAGE [{message_id}]")
    nem = imap_agent.fetch_mail_with_message_id_as_nem(message_id)
    print(f'\tOn {nem.read_field_date()}')
    print(f'\tFrom {nem.read_field_from()} To {nem.read_field_to()}')
    print(f'\tSubject {nem.read_field_subject()}')

imap_agent.logout()



import csv
from typing import Optional, Union, Type


class CSVReaderOptions:
    def __init__(
            self,
            dialect: Optional[Union[csv.Dialect, Type[csv.Dialect], str]] = None,
            delimiter: Optional[str] = None,
            doublequote: Optional[bool] = None,
            escapechar: Optional[str] = None,
            lineterminator: Optional[str] = None,
            quotechar: Optional[str] = None,
            quoting: Optional[int] = None,
            skipinitialspace: Optional[bool] = None,
            strict: Optional[bool] = None
    ):
        self.dialect = dialect
        self.delimiter = delimiter
        self.doublequote = doublequote
        self.escapechar = escapechar
        self.lineterminator = lineterminator
        self.quotechar = quotechar
        self.quoting = quoting
        self.skipinitialspace = skipinitialspace
        self.strict = strict

        if dialect is not None:
            self.dialect = dialect
        if delimiter is not None:
            self.delimiter = delimiter
        if doublequote is not None:
            self.doublequote = doublequote
        if escapechar is not None:
            self.escapechar = escapechar
        if lineterminator is not None:
            self.lineterminator = lineterminator
        if quotechar is not None:
            self.quotechar = quotechar
        if quoting is not None:
            self.quoting = quoting
        if skipinitialspace is not None:
            self.skipinitialspace = skipinitialspace
        if strict is not None:
            self.strict = strict

    def to_dict(self):
        reader_kwargs = dict()
        if self.dialect is not None:
            reader_kwargs['dialect'] = self.dialect
        if self.delimiter is not None:
            reader_kwargs['delimiter'] = self.delimiter
        if self.doublequote is not None:
            reader_kwargs['doublequote'] = self.doublequote
        if self.escapechar is not None:
            reader_kwargs['escapechar'] = self.escapechar
        if self.lineterminator is not None:
            reader_kwargs['lineterminator'] = self.lineterminator
        if self.quotechar is not None:
            reader_kwargs['quotechar'] = self.quotechar
        if self.quoting is not None:
            reader_kwargs['quoting'] = self.quoting
        if self.skipinitialspace is not None:
            reader_kwargs['skipinitialspace'] = self.skipinitialspace
        if self.strict is not None:
            reader_kwargs['strict'] = self.strict

        return reader_kwargs



from nehushtan.MessageQueue.implement.NehushtanMemoryMessageQueue import NehushtanMemoryMessageQueue
from nehushtan.socket.NehushtanUDPSocketServer import NehushtanUDPSocketServer
from tests.socket.badmq.MQCommand import MQCommand


class BadMQServer(NehushtanUDPSocketServer):

    def __init__(self, host: str, port: int):
        super().__init__(host, port)

        self.__mq = NehushtanMemoryMessageQueue()

    def handle_incoming_data(self, data: bytes, address):
        print(f'RECV: {data}')
        x = MQCommand.parse(data)
        if x:
            if x.command == MQCommand.COMMAND_DEQUEUE:
                x.value = self.__mq.dequeue(x.queue_name)
                response = x.package_as_bytes()
            elif x.command == MQCommand.COMMAND_ENQUEUE:
                x.value = self.__mq.enqueue(x.value, x.queue_name)
                response = x.package_as_bytes()
            else:
                response = '{"ERROR":"UNKNOWN COMMAND"}'.encode()
        else:
            response = '{"ERROR":"FORMAT ERROR"}'.encode()

        print(f"SEND: {response}")
        self.get_socket_instance().sendto(response, address)


if __name__ == '__main__':
    BadMQServer('127.0.0.1', 6767).listen()



import socketserver
from typing import Tuple, Sequence, Union

from nehushtan.MessageQueue.implement.NehushtanMemoryMessageQueue import NehushtanMemoryMessageQueue
from nehushtan.httpd.NehushtanHTTPRequestHandler import NehushtanHTTPRequestHandler
from nehushtan.httpd.NehushtanHTTPRouter import NehushtanHTTPRouter


class TestRequestHandler(NehushtanHTTPRequestHandler):
    router = NehushtanHTTPRouter()
    mq = NehushtanMemoryMessageQueue()

    def __init__(self, request, client_address: Tuple[str, int], server: socketserver.BaseServer):
        # self.prepare_router()
        super().__init__(request, client_address, server)

    def seek_route_for_process_chains(self, method: str, path: str) \
            -> Tuple[Sequence[Union[type, str]], Tuple[Union[type, str], str]]:
        route = TestRequestHandler.router.check_request_for_route(method, path)
        # self.log_message(route.path_template)
        self.matched_arguments = route.matched_arguments
        self.matched_keyed_arguments = route.matched_keyed_arguments
        return route.get_filter_list(), route.get_controller_target()



from nehushtan.httpd.NehushtanHTTPRequestController import NehushtanHTTPRequestController


class TestProcessChainB(NehushtanHTTPRequestController):
    def list(self, x: str):
        index = self._get_http_handler().parsed_query_dict.get('index', -1)
        # header_cookie=self._read_header('cookie')
        a = self._read_cookie('a')
        b = self._read_cookie(u'喵')
        self._reply_with_ok(
            {
                "index": index,
                'cookie': self._get_http_handler().parsed_cookie_dict,
                'a': a,
                'b': b,
                'x': x,
            }
        )

    def step3(self, step: float):
        x = self._get_http_handler().get_process_chain_share_data_dict().get('index', step)
        self._reply_with_ok(x)

    def mix(self, index: int, name: str, value: float, action: str):
        self._reply_with_ok({'index': index, 'name': name, 'value': value, 'action': action})

    def args(self, a: int, b: float, c: str, d: str = 'd', e: str = 'e'):
        self._reply_with_ok({'a': a, 'b': b, 'c': c, 'd': d, 'e': e})



#  Copyright (c) 2020. Sinri Edogawa
from typing import Iterable, Optional, Union, List

from nehushtan.mysql.MySQLCondition import MySQLCondition
from nehushtan.mysql.MySQLKit import MySQLKit
from nehushtan.mysql.MySQLSelectionMixin import MySQLSelectionMixin, MySQLSelectionTarget
from nehushtan.mysql.MySQLTableExistence import MySQLTableExistence





class MySQLTableSelection(MySQLSelectionMixin):
    """
    As of version 0.3.6, move some features to MySQLSelectionMixin
    """

    def __init__(self,
                 target: MySQLSelectionTarget,
                 # model: MySQLTableExistence, alias: Optional[str] = None,
                 ):
        super().__init__()
        # self._model = model
        # self._alias = alias
        self._target = target
        self._select_fields = []
        # self._join_metas = []
        self._conditions = []
        self._group_by_fields = []
        self._sort_expression = ''
        self._limit = 0
        self._offset = 0
        self._use_indices = []
        self._force_indices = []
        self._ignore_indices = []
        self._for_update = False

    def get_mysql_kit(self) -> MySQLKit:
        return self._target.get_mysql_kit()

    def get_limit(self) -> int:
        return self._limit

    def set_limit(self, limit: int):
        self._limit = limit
        return self

    def get_offset(self) -> int:
        return self._offset

    def set_offset(self, offset: int):
        self._offset = offset
        return self

    def get_sort_expression(self):
        return self._sort_expression

    def set_sort_expression(self, sort_expression: str):
        self._sort_expression = sort_expression
        return self

    def add_select_field(self, field_expression: str, alias: str = ''):
        s = field_expression
        if len(alias) > 0:
            s += f' as {alias}'
        self._select_fields.append(s)
        return self

    def add_select_field_name_list(self, field_name_array: Iterable[str]):
        """

        :param field_name_array: ["F1","F2", ...] or ("F1","F2", ...)
        :return:
        """
        for item in field_name_array:
            self._select_fields.append(item)
        return self

    # def add_left_join_table(self, table_name: str, schema_name: Optional[str] = None,
    #                         alias: Optional[str] = None,
    #                         on_conditions: Optional[Iterable[MySQLJoinOnCondition]] = None, ):
    #     return self._add_join_table(join_type="LEFT JOIN", table_name=table_name, schema_name=schema_name, alias=alias,
    #                                 on_conditions=on_conditions)
    #
    # def add_inner_join_table(self, table_name: str, schema_name: Optional[str] = None,
    #                          alias: Optional[str] = None,
    #                          on_conditions: Optional[Iterable[MySQLJoinOnCondition]] = None, ):
    #     return self._add_join_table(join_type="INNER JOIN", table_name=table_name, schema_name=schema_name, alias=alias,
    #                                 on_conditions=on_conditions)

    # def _add_join_table(self, join_type: str, table_name: str, schema_name: Optional[str] = None,
    #                     alias: Optional[str] = None, on_conditions: Optional[Iterable[MySQLJoinOnCondition]] = None, ):
    #     e = f'`{table_name}`'
    #     if schema_name is not None:
    #         e = f'`{schema_name}`.{e}'
    #     if alias is not None:
    #         e = f'{e} as `{alias}`'
    #     c = ''
    #     if on_conditions is not None:
    #         c_array = []
    #         for condition in on_conditions:
    #             c_array.append(f'({condition.organize_to_sql()})')
    #         c_c = ' AND '.join(c_array)
    #         c = f' ON {c_c}'
    #
    #     j = f'{join_type} {e}{c}'
    #
    #     self._join_metas.append(j)
    #     return self

    # def add_left_join_sub_query(self, selection: MySQLSelectionMixin, alias: str,
    #                             on_conditions: Optional[Iterable[MySQLJoinOnCondition]] = None, ):
    #     return self._add_join_sub_query("LEFT JOIN", selection=selection, alias=alias, on_conditions=on_conditions)
    #
    # def add_inner_join_sub_query(self, selection: MySQLSelectionMixin, alias: str,
    #                              on_conditions: Optional[Iterable[MySQLJoinOnCondition]] = None, ):
    #     return self._add_join_sub_query("INNER JOIN", selection=selection, alias=alias, on_conditions=on_conditions)
    #
    # def _add_join_sub_query(self, join_type: str, selection: MySQLSelectionMixin, alias: str,
    #                         on_conditions: Optional[Iterable[MySQLJoinOnCondition]] = None, ):
    #     c = ''
    #     if on_conditions is not None:
    #         c_array = []
    #         for condition in on_conditions:
    #             c_array.append(f'({condition.organize_to_sql()})')
    #         c_c = ' AND '.join(c_array)
    #         c = f' ON {c_c}'
    #     j = f'{join_type} ({selection.generate_sql()}) as `{alias}`{c}'
    #     self._join_metas.append(j)
    #     return self

    def add_condition(self, condition: MySQLCondition):
        self._conditions.append(condition)
        return self

    def add_conditions(self, conditions: Iterable[MySQLCondition]):
        """

        :param conditions: array of MySQLCondition
        :return:
        """
        for condition in conditions:
            if isinstance(condition, MySQLCondition):
                self._conditions.append(condition)
            # else just ignore
        return self

    def add_simple_conditions(self, equal_dict: dict):
        """

        :param equal_dict: {FIELD:VALUE,FIELD:VALUE_ARRAY}
        :return:
        """
        for k, v in equal_dict.items():
            if isinstance(v, (tuple, list)):
                self.add_condition(MySQLCondition.make_in_array(k, v))
            else:
                self.add_condition(MySQLCondition.make_equal(k, v))
        return self

    def set_group_by_fields(self, group_by_fields: list):
        self._group_by_fields = group_by_fields
        return self

    def use_index(self, index: str):
        self._use_indices.append(index)
        return self

    def force_index(self, index: str):
        self._force_indices.append(index)
        return self

    def ignore_index(self, index: str):
        self._ignore_indices.append(index)
        return self

    def set_for_update(self, value: bool):
        self._for_update = value

    def generate_sql(self) -> str:
        # table = self._model.get_table_expression()
        #
        # if self._alias is not None:
        #     table += f' AS `{self._alias}` '

        target = self._target.as_selection_target()

        fields = "*"
        if len(self._select_fields) > 0:
            fields = ",".join(self._select_fields)

        condition_sql = MySQLCondition.build_sql_component(self._conditions)

        if isinstance(self._target, MySQLTableExistence):
            indices = ''
            if len(self._use_indices) > 0:
                indices += " USE INDEX (" + ",".join(self._use_indices) + ") "
            if len(self._force_indices) > 0:
                indices += " FORCE INDEX (" + ",".join(self._force_indices) + ") "
            if len(self._ignore_indices) > 0:
                indices += " IGNORE INDEX (" + ",".join(self._force_indices) + ") "
            target += indices

        # joins = ' '.join(self._join_metas)

        # sql = f"SELECT {fields} FROM {table} {indices} {joins} WHERE {condition_sql} "

        sql = f"SELECT {fields} FROM {target} WHERE {condition_sql} "

        if len(self._group_by_fields) > 0:
            sql += "GROUP BY " + ",".join(self._group_by_fields) + " "

        if self._sort_expression.strip() != '':
            sql += "ORDER BY " + self._sort_expression + " "

        if self._limit > 0:
            sql += f"LIMIT {self._limit} "
            if self._offset > 0:
                sql += f"OFFSET {self._offset} "

        if self._for_update:
            sql += " FOR UPDATE "

        return sql



from nehushtan.helper.CommonHelper import CommonHelper
from tests.postgres.PgSQLTestEnv import PgSQLTestEnv


class TestBasic1:
    def __init__(self):
        self.__env = PgSQLTestEnv()
        self.__kit = self.__env.get_pg_kit_instance()

    def test(self):
        # self.__create_schema()
        # self.__create_table_1()
        self.__insert_some()
        self.__query_all()
        self.__query_some_1()
        self.__query_some_2()
        self.__done_transcation()
        self.__undone_transcation()
        # self.__drop_schema_cascade()

    def __create_schema(self):
        result = self.__kit.execute('create schema nehushtan_test_basic')
        self.__env.log_for_result_set(result)

    def __create_table_1(self):
        result = self.__kit.execute(
            '''create table nehushtan_test_basic.table_1 (
                id serial,
                name varchar(255) COLLATE "pg_catalog"."default",
                content varchar(255),
                act_time timestamp,
                primary key (id)
            );'''
        )
        self.__env.log_for_result_set(result)

    def __insert_some(self):
        content = CommonHelper.generate_a_password_string(16)
        result = self.__kit.execute(
            """insert into nehushtan_test_basic.table_1(name,content,act_time)values(%s,%s,now()) returning id""",
            ('a', content)
        )
        self.__env.log_for_result_set(result)

    def __query_all(self):
        result = self.__kit.execute(
            """select * from nehushtan_test_basic.table_1"""
        )
        self.__env.log_for_result_set(result)

    def __query_some_1(self):
        result = self.__kit.execute(
            """select 
                name,to_timestamp(avg(EXTRACT(epoch FROM act_time)))
            from nehushtan_test_basic.table_1
            where name='a' 
            group by name"""
        )
        self.__env.log_for_result_set(result)

    def __query_some_2(self):
        result = self.__kit.execute(
            """select
                id,name,content,'do not use value binding, one percent char for one'
            from nehushtan_test_basic.table_1
            where content like '%a%'
            """
        )
        self.__env.log_for_result_set(result)
        result = self.__kit.execute(
            """select
                id,name,content,'use value binding, two percent chars for one'
            from nehushtan_test_basic.table_1
            where content like '%%a%%'
            and id = %s
            """,
            (7,)
        )
        self.__env.log_for_result_set(result)

        result = self.__kit.execute(
            """select
                id,name,content,'use value binding, just put percent into value'
            from nehushtan_test_basic.table_1
            where content like %s
            """,
            ('%a%',)
        )
        self.__env.log_for_result_set(result)

    def __done_transcation(self):
        with self.__kit.get_raw_connection():
            content = CommonHelper.generate_a_password_string(16)
            result = self.__kit.execute(
                """insert into nehushtan_test_basic.table_1(name,content,act_time)
                values(%s,%s,now()) 
                returning id""",
                ('transcation_done', content)
            )
            self.__env.log_for_result_set(result)

    def __undone_transcation(self):
        try:
            with self.__kit.get_raw_connection():
                content = CommonHelper.generate_a_password_string(16)
                result = self.__kit.execute(
                    """insert into nehushtan_test_basic.table_1(name,content,act_time)
                    values(%s,%s,now()) 
                    returning id""",
                    ('transcation_undone', content)
                )
                self.__env.log_for_result_set(result)
                raise Exception('let us rollback!')
        except:
            self.__env.log('ROLLBACK')

    def __drop_schema_cascade(self):
        result = self.__kit.execute('drop schema nehushtan_test_basic CASCADE')
        self.__env.log_for_result_set(result)


if __name__ == '__main__':
    x = TestBasic1()
    x.test()



from nehushtan.logger.NehushtanLogger import NehushtanLogger
from nehushtan.mysql import constant
from nehushtan.mysql.MySQLKit import MySQLKit
from nehushtan.mysql.MySQLKitConfig import MySQLKitConfig
from tests.config import MYSQL_CONFIG

if __name__ == '__main__':
    logger = NehushtanLogger('default')

    config = MySQLKitConfig(MYSQL_CONFIG)
    db = MySQLKit(config)

    # print(db.get_raw_connection().cursor().lastrowid)
    # exit(0)

    replace_sql = 'replace into c(id,value,name) values(%s,%s,%s)'

    x = db.raw_query_to_insert_one(replace_sql, (1, 1, 'a'))
    logger.info('Round 1', x)

    x = db.raw_query_to_insert_one(replace_sql, (2, 2, 'a'))
    logger.info('Round 2', x)

    x = db.raw_query_to_insert_one(replace_sql, (2, 2, 'a'))
    logger.info('Round 3', x)

    x = db.raw_query_to_insert_one(replace_sql, [])
    logger.info('Round 4', x)

    cursor = db.get_raw_connection().cursor()

    cursor.executemany(replace_sql, [[]])

    print(constant.MYSQL_QUERY_STATUS_EXECUTED)
    print(cursor.lastrowid)
    print(cursor.rowcount)



class NehushtanWebsocketConnectionEntity:
    def __init__(self, websocket):
        self.__local_key = f'{websocket.local_address[0]}-{websocket.local_address[1]}'
        self.__remote_key = f'{websocket.remote_address[0]}-{websocket.remote_address[1]}'
        self.__websocket = websocket

    def get_local_key(self):
        return self.__local_key

    def get_remote_key(self):
        return self.__remote_key

    def get_key(self):
        return f'{self.__local_key}~{self.__remote_key}'

    def get_websocket(self):
        return self.__websocket



import json
import time
from typing import List

from aliyun.log import LogClient, PutLogsRequest, LogItem, LogException

from nehushtan.helper.CommonHelper import CommonHelper
from nehushtan.logger.NehushtanLogger import NehushtanLoggerAdapter, NehushtanLogLevel


class AliyunSLSLoggerAdapter(NehushtanLoggerAdapter):
    """
    Adapter for Aliyun SLS.
    Requirement `aliyun-log-python-sdk` (such as 0.9.11) should be declared.
    Since 0.5.1.
    """

    def __init__(self,
                 endpoint: str,
                 access_key_id: str,
                 access_key_secret: str,
                 project: str,
                 logstore: str,
                 source: str,
                 with_buffer: bool = False,
                 buffer_size: int = 100,
                 ):
        super().__init__()
        self.__client = LogClient(endpoint, access_key_id, access_key_secret)
        self.__logstore = logstore
        self.__project = project
        self.__source = source.replace('[IP]', CommonHelper.get_local_ip())
        self.__with_buffer = with_buffer
        self.__buffer = []
        self.__buffer_size = buffer_size

    def write_one_log(self, level: NehushtanLogLevel, contents: dict):
        if self.__with_buffer:
            return self.__write_one_log_into_buffer(level, contents)
        else:
            return self.__write_one_log_directly(level, contents)

    def __write_one_log_directly(self, level: NehushtanLogLevel, contents: dict):
        # topic = contents['topic']
        # del contents['topic']
        contents['level'] = level.name

        l = []
        for (k, v) in contents.items():
            if v is None:
                l.append((k, None,))
            elif isinstance(v, str):
                l.append((k, v))
            else:
                l.append((k, json.dumps(v, default=lambda x: x.__str__(), ensure_ascii=False)))

        log_item = LogItem()
        log_item.set_time(int(time.time()))
        log_item.set_contents(l)

        self.__send_to_sls([log_item])

    def __write_one_log_into_buffer(self, level: NehushtanLogLevel, contents: dict):
        contents['level'] = level.name
        self.__buffer.append(contents)

        if len(self.__buffer) >= self.__buffer_size:
            self.flush_buffer()

    def flush_buffer(self):
        if len(self.__buffer) == 0:
            return

        log_item_list = []
        for contents in self.__buffer:
            l = []
            for (k, v) in contents.items():
                if v is None:
                    l.append((k, None,))
                elif isinstance(v, str):
                    l.append((k, v))
                else:
                    l.append((k, json.dumps(v, default=lambda x: x.__str__(), ensure_ascii=False)))

            log_item = LogItem()
            log_item.set_time(int(time.time()))
            log_item.set_contents(l)

            log_item_list.append(log_item)

        self.__send_to_sls(log_item_list)

    def __send_to_sls(self, log_item_list: List[LogItem]):
        req2 = PutLogsRequest(
            project=self.__project,
            logstore=self.__logstore,
            topic=self.get_topic(),
            source=self.__source,
            logitems=log_item_list,
        )
        try:
            self.__client.put_logs(req2)
        except LogException as e:
            print(e)

    def __del__(self):
        if self.__with_buffer and len(self.__buffer) > 0:
            self.flush_buffer()



#  Copyright (c) 2020. Sinri Edogawa
from typing import Callable

import pymysql
from pymysql.connections import Connection
from pymysql.cursors import DictCursor

from nehushtan.logger.NehushtanLogger import NehushtanLogLevel, NehushtanLogger
from nehushtan.mysql import constant
from nehushtan.mysql.MySQLKitConfig import MySQLKitConfig


class MySQLKit:
    """
    MySQL 连接基础工具类，封装了 PYMYSQL 库的实例。
    Shovel 项目相关的所有 MySQL 连接应当使用此方法进行。
    Since 0.1.13 Added Auto Reconnection (PING)
    """

    # _mysql_config: MySQLKitConfig
    # _auto_commit_default: bool
    # _connection: Connection or None

    def __init__(self, mysql_config: MySQLKitConfig = None):
        """
        Since 0.1.4 入参类型变化
        获取一个MySQLKitConfig类的实例，并根据此类设置的配置立即建立一个新的数据库连接。
        PYMYSQL 默认不自动递交。在实际场景中，建议设置成默认自动递交并仅在需要时手动开启事务。
        :param mysql_config: 获取一个MySQLKitConfig类的实例，可用字典构建，样本参见配置文件标准示例中 `mysql_database`.`sample` 项
        """
        # Planned in 0.3.6 but not confirmed
        # self.__instance_uuid = uuid.uuid1().urn

        if mysql_config is None:
            mysql_config = MySQLKitConfig()
        self._mysql_config = mysql_config
        self._auto_commit_default = self._mysql_config.get_auto_commit()
        self._connection = None
        if len(self._mysql_config.get_host()) > 0:
            self.connect()

        self.__logger = NehushtanLogger(log_level=NehushtanLogLevel.FATAL)

    def set_logger(self, logger: NehushtanLogger):
        self.__logger = logger

    def get_logger(self) -> NehushtanLogger:
        return self.__logger

    def __del__(self):
        """
        Since 0.4.10
        """
        self.disconnect()

    # Planned in 0.3.6 but not confirmed
    # def get_mysql_kit_uuid(self):
    #     return self.__instance_uuid

    @staticmethod
    def make_instance_from_pymysql_connection(connection: Connection):
        """
        根据一个已经存在的 PYMYSQL 连接，构建一个本类实例。
        这样建立的实例里的配置字典为空，而自动提交参数根据入参连接的实际情况获取。
        :param connection:
        :return: 本类实例
        """
        kit = MySQLKit()
        kit._connection = connection
        kit._auto_commit_default = connection.get_autocommit()
        return kit

    def get_raw_connection(self) -> Connection:
        """
        :return: 封装的 PYMYSQL 原生 CONNECTION 实例
        """
        if self._connection is None:
            raise RuntimeError("Connection Property of MySQLKit Instance is not set.")
        self._connection.ping(reconnect=True)
        return self._connection

    def connect(self):
        """
        It may raise Exception if connection failed, as I viewed the document
        :return: MySQLKit the instance itself
        """

        self.disconnect()

        self._auto_commit_default = self._mysql_config.get_auto_commit()
        self._connection = pymysql.connect(
            host=self._mysql_config.get_host(),  # '172.16.1.52',
            port=self._mysql_config.get_port(),  # 3306,
            user=self._mysql_config.get_user(),
            password=self._mysql_config.get_password(),
            db=self._mysql_config.get_db(),
            charset=self._mysql_config.get_charset(),
            autocommit=self._auto_commit_default,
        )
        return self

    def disconnect(self):
        if self._connection is not None:
            self._connection.close()
        self._connection = None
        return self

    def raw_get_computed_sql(self, sql: str, args=None) -> str:
        """
        使用 PYMYSQL 原生的方法获取构建的 SQL 语句用于预览
        :param sql: SQL 模板
        :param args: 参数
        :return: 生成的 SQL 语句
        """
        connection = self.get_raw_connection()
        cursor = connection.cursor()
        sql = cursor.mogrify(sql, args)
        cursor.close()
        return sql

    def raw_query_for_all_tuple_rows(self, sql: str, args=None) -> tuple:
        """
        查询一个SQL并获取包含所有结果行的元组，其中的每行也封装成一个元组
        :param sql:
        :param args:
        :return:
        """
        connection = self.get_raw_connection()
        cursor = connection.cursor()

        self.get_logger().info(f'nehushtan.mysql.MySQLKit.MySQLKit.raw_query_for_all_tuple_rows to go', {
            'sql': sql, 'args': args,
        })

        cursor.execute(sql, args)
        rows_tuple = cursor.fetchall()
        cursor.close()
        return rows_tuple

    def raw_query_for_all_dict_rows(self, sql: str, args=None) -> tuple:
        """
        查询一个SQL并获取包含所有结果行的元组，其中的每行封装成一个字典
        :param sql:
        :param args:
        :return:
        """
        connection = self.get_raw_connection()
        cursor = connection.cursor(DictCursor)

        self.get_logger().info(f'nehushtan.mysql.MySQLKit.MySQLKit.raw_query_for_all_dict_rows to go', {
            'sql': sql, 'args': args,
        })

        cursor.execute(sql, args)
        rows_tuple = cursor.fetchall()
        cursor.close()
        return rows_tuple

    def raw_query_to_modify_one(self, sql: str, args=None, commit_immediately: bool = None):
        """
        执行一个非读SQL并返回影响行数
        :param sql:
        :param args:
        :param commit_immediately:
        :return:
        """
        connection = self.get_raw_connection()
        cursor = connection.cursor()

        self.get_logger().info(f'nehushtan.mysql.MySQLKit.MySQLKit.raw_query_to_modify_one to go', {
            'sql': sql, 'args': args,
        })

        afx = cursor.execute(sql, args)
        if commit_immediately is None:
            commit_immediately = self._auto_commit_default
        if commit_immediately:
            connection.commit()
        cursor.close()
        return afx

    def raw_query_to_insert_one(self, sql: str, args=None, commit_immediately: bool = None):
        """
        执行一条插入语句并获取最后更新的行的ID
        :param sql:
        :param args:
        :param commit_immediately:
        :return:
        """
        connection = self.get_raw_connection()
        cursor = connection.cursor()

        self.get_logger().info(f'nehushtan.mysql.MySQLKit.MySQLKit.raw_query_to_insert_one to go', {
            'sql': sql, 'args': args,
        })

        cursor.execute(sql, args)
        last_row_id = cursor.lastrowid
        if commit_immediately is None:
            commit_immediately = self._auto_commit_default
        if commit_immediately:
            connection.commit()
        cursor.close()
        return last_row_id

    def raw_query_to_modify_many(self, sql: str, args=None, commit_immediately: bool = None):
        """
        执行原生的SQL批量变更并获取影响行数
        :param sql:
        :param args:
        :param commit_immediately:
        :return:
        """
        connection = self.get_raw_connection()
        cursor = connection.cursor()

        self.get_logger().info(f'nehushtan.mysql.MySQLKit.MySQLKit.raw_query_to_modify_many to go', {
            'sql': sql, 'args': args,
        })

        afx = cursor.executemany(sql, args)
        if commit_immediately is None:
            commit_immediately = self._auto_commit_default
        if commit_immediately:
            connection.commit()
        cursor.close()
        return afx

    def raw_query_to_insert_many(self, sql: str, args=None, commit_immediately: bool = None):
        """
        执行原生的SQL批量插入并获取最后更新行的ID
        :param sql:
        :param args:
        :param commit_immediately:
        :return:
        """
        connection = self.get_raw_connection()
        cursor = connection.cursor()

        self.get_logger().info(f'nehushtan.mysql.MySQLKit.MySQLKit.raw_query_to_insert_many to go', {
            'sql': sql, 'args': args,
        })

        cursor.executemany(sql, args)
        last_row_id = cursor.lastrowid
        if commit_immediately is None:
            commit_immediately = self._auto_commit_default
        if commit_immediately:
            connection.commit()
        cursor.close()
        return last_row_id

    def raw_execute_transaction(self, transaction_callable: Callable[["MySQLKit"], any]):
        """
        一个封装的方法，用于快速包装事务。
        :param transaction_callable: 事务处理具体代码所在的闭包，可以使用lambda表达式或者def方法，参数为本类实例，并返回结果
        :return: 返回闭包的返回值
        """
        connection = self.get_raw_connection()
        connection.begin()
        try:
            result = transaction_callable(self)
            self.get_logger().notice(f'to commit with result {result}')
            connection.commit()
            return result
        except Exception as e:
            self.get_logger().notice(f'to rollback with exception {e}')
            connection.rollback()
            raise e

    def quote(self, value):
        """
        将任意入参进行转义以防范注入。
        如果连接关闭，则使用脱机转义。
        数字、空值、布尔值不加引号，其他都作为字符串加外部引号可用于字符串拼接。
        如果入参为元组或者列表，也以同样形式返回转义所有元素的新实例。
        :param value:
        :return:
        """
        if self._connection is None:
            return self.quote_offline(value)
        self._connection.ping(True)

        if type(value) == list:
            x = []
            for item in value:
                x.append(self.quote(item))
            return x
        elif type(value) == tuple:
            x = []
            for item in value:
                x.append(self.quote(item))
            return tuple(x)
        elif type(value) == int or type(value) == float:
            # Fixed in 0.1.8: if number is returned, the str plus would fail
            return f"{value}"
        elif value is None:
            return constant.MYSQL_CONDITION_CONST_NULL
        elif value is True:
            return constant.MYSQL_CONDITION_CONST_TRUE
        elif value is False:
            return constant.MYSQL_CONDITION_CONST_FALSE
        elif type(value) == str:
            escaped = self._connection.escape_string(value)
            return f"'{escaped}'"
        else:
            return self.quote(value.__str__())

    @staticmethod
    def quote_offline(value):
        """
        脱机转义
        :param value:
        :return:
        """
        if type(value) == list:
            x = []
            for item in value:
                x.append(MySQLKit.quote_offline(item))
            return x
        elif type(value) == tuple:
            x = []
            for item in value:
                x.append(MySQLKit.quote_offline(item))
            return tuple(x)
        elif type(value) == int or type(value) == float:
            return f'{value}'
        elif type(value) == str:
            return MySQLKit.quote_string_offline(value)
        elif value is None:
            return constant.MYSQL_CONDITION_CONST_NULL
        elif value is True:
            return constant.MYSQL_CONDITION_CONST_TRUE
        elif value is False:
            return constant.MYSQL_CONDITION_CONST_FALSE
        else:
            return MySQLKit.quote_string_offline(value.__str__())

    @staticmethod
    def quote_string_offline(value: str) -> str:
        """
        脱机转义字符串。
        :param value:
        :return: 包含了外层引号
        """
        d = {
            '\\': '\\\\',
            '\0': '\\0',
            '\n': '\\n',
            '\r': '\\r',
            "'": "\\'",
            '"': '\\"',
            '\x1a': '\\Z',
        }
        s = value
        for k, v in d.items():
            s = s.replace(k, v)

        return f"'{s}'"



import uuid
from typing import Dict, List

from nehushtan.ws.NehushtanWebsocketConnectionEntity import NehushtanWebsocketConnectionEntity


class TestWebsocketRegisterAgent:

    def __init__(self):
        self.__map: Dict[str, NehushtanWebsocketConnectionEntity] = {}
        self.agent_identity = str(uuid.uuid4())

    def register(self, websocket):
        entity = NehushtanWebsocketConnectionEntity(websocket)
        self.__map[entity.get_key()] = entity
        print(f"TestWebsocketRegisterAgent[{self.agent_identity}] registered [{entity.get_key()}]")
        return entity

    def unregister(self, key: str):
        if self.__map.get(key):
            del self.__map[key]
            print(f"TestWebsocketRegisterAgent[{self.agent_identity}] unregistered [{key}]")

    def read(self, key: str):
        print(f"TestWebsocketRegisterAgent[{self.agent_identity}] reading [{key}]")
        return self.__map.get(key)

    def list_for_server(self, local_key: str) -> List[NehushtanWebsocketConnectionEntity]:
        print(f"TestWebsocketRegisterAgent[{self.agent_identity}] listing for [{local_key}]")
        enities = []
        for k, v in self.__map.items():
            if v.get_local_key() == local_key:
                enities.append(v)
        return enities



import socket
from abc import abstractmethod
from threading import Thread
from typing import Optional

from nehushtan.socket.SocketHandleThreadManager import SocketHandlerThreadManager


class NehushtanTCPSocketServer:
    """
    Since 0.4.16
    """

    def __init__(self, port: int, host=None, wait_queue_size=0, tm: SocketHandlerThreadManager = None):
        self.__listen_port = port
        self.__listen_host = host
        self.__socket_instance: Optional[socket.socket] = None
        self.__wait_queue_size = wait_queue_size
        if not tm:
            tm = SocketHandlerThreadManager()
        self.__thread_manager = tm

    def get_thread_manager(self):
        return self.__thread_manager

    def listen(self):
        address_entries = socket.getaddrinfo(
            self.__listen_host,
            self.__listen_port,
            socket.AF_UNSPEC,
            socket.SOCK_STREAM,
            0,
            socket.AI_PASSIVE
        )

        prepare_error = None

        for address_entry in address_entries:
            af, socket_type, proto, canon_name, sa = address_entry
            try:
                self.__socket_instance = socket.socket(af, socket_type, proto)
                self.__socket_instance.bind(sa)
                self.__socket_instance.listen(self.__wait_queue_size)
            except OSError as msg:
                prepare_error = msg
                if self.__socket_instance is not None:
                    self.__socket_instance.close()
                self.__socket_instance = None
                continue
            break

        if self.__socket_instance is None:
            raise prepare_error

        # let the process of accept could be interrupted (Since 0.4.17)
        self.__socket_instance.settimeout(1)

        while not self.should_terminate():
            try:
                connection, address = self.__socket_instance.accept()
            except socket.timeout:
                continue

            thread = Thread(
                target=self.handle_incoming_connection,
                name=connection.getpeername(),
                args=(connection, address)
            )
            self.__thread_manager.register_thread(thread)
            thread.start()

    def should_terminate(self) -> bool:
        """
        If return True, stop the listener.
        Override it to implement it with your own logic.
        """
        return False

    @abstractmethod
    def handle_incoming_connection(self, connection: socket.socket, address):
        """
        IN THREAD.
        Override it to implement it with your own logic.
        You may need to close client connection inside
        """
        pass



import plotly.express as px

if __name__ == '__main__':
    fig = px.bar(x=["a", "b", "c"], y=[1, 3, 2])
    fig.write_html('/Users/leqee/code/nehushtan/debug/plotly/first_figure.html', auto_open=True)



from typing import List, Optional

from nehushtan.paint.Paint import Paint
from nehushtan.score.ParsedLine import ParsedLine
from nehushtan.score.ParsedLineAsBlank import ParsedLineAsBlank
from nehushtan.score.ParsedLineAsLyric import ParsedLineAsLyric
from nehushtan.score.ParsedLineAsScore import ParsedLineAsScore
from nehushtan.score.ParsedLineAsTitle import ParsedLineAsTitle
from nehushtan.score.ScoreDrawerOptions import ScoreDrawerOptions


class ScoreDrawer:
    def __init__(self,
                 parsed_lines: List[ParsedLine],
                 options: ScoreDrawerOptions
                 ):
        self.__parsed_lines = parsed_lines

        self.__options = options

        cell_size_param = self.__options.get_default_font_size()

        self.__line_width = int(max(3.0, cell_size_param / 12))
        self.__point_width = int(max(3.0, cell_size_param / 12))

        self.__max_columns = 0
        self.__max_prefix_columns = 0
        self.__max_score_columns = 0

        self.__total_cells_as_width = 0
        self.__total_cells_as_height = 0

        self.__compute_cells()

        self.__cell_width = cell_size_param
        self.__cell_height = cell_size_param
        self.__paint: Optional[Paint] = None

        self.__prepare_image_paper()

        self.__draw_lines()

    def show(self, title: Optional[str] = None):
        self.__paint.show(title)

    def save(self, output_file_path):
        self.__paint.save(output_file_path, "png")

    def __compute_cells(self):
        max_columns = 0
        max_prefix_columns = 0
        max_score_columns = 0
        for line in self.__parsed_lines:
            if isinstance(line, ParsedLineAsScore):
                # compute for score
                cells = 0
                for su in line.get_score_units():
                    cells += su.get_cell_needed()
                if cells > max_score_columns:
                    max_score_columns = cells
            elif isinstance(line, ParsedLineAsLyric):
                prefix_size = 0
                if line.get_prefix() is not None:
                    prefix_size += len(line.get_prefix())  # + 1

                if prefix_size > max_prefix_columns:
                    max_prefix_columns = prefix_size

                total_chars = 0
                total_chars += len(line.get_content())
                if total_chars > max_columns:
                    max_columns = total_chars
            elif isinstance(line, ParsedLineAsTitle):
                # do not care title?
                if False:
                    total_chars = len(line.get_middle_component())
                    if line.get_left_component() is not None:
                        total_chars += 1 + len(line.get_left_component())
                    if line.get_right_component() is not None:
                        total_chars += 1 + len(line.get_right_component())
                    if int(total_chars * 0.8) > max_columns:
                        max_columns = int(total_chars * 0.8)

        self.__max_columns = max_columns
        self.__max_prefix_columns = max_prefix_columns
        self.__max_score_columns = max_score_columns

        print(f"self.__max_columns = {self.__max_columns}")
        print(f"self.__max_prefix_columns = {self.__max_prefix_columns}")
        print(f"self.__max_score_columns = {self.__max_score_columns}")

    def __prepare_image_paper(self):
        with Paint(400, 400) as p:
            x = p.textbbox(
                (0, 0),
                "喵",
                self.__options.get_default_font(),
                align="center"
            )
            # print(x)
            self.__cell_width = x[2] + 2
            self.__cell_height = int((x[3] + 2) * 1.5)
            # print(f'self.__cell_width={self.__cell_width} self.__cell_height={self.__cell_height}')

        if self.__options.get_total_cells_in_one_line() is not None:
            self.__total_cells_as_width = (2 + self.__options.get_total_cells_in_one_line())
            print(f"width in cells: {self.__total_cells_as_width}")
            width = self.__cell_width * self.__total_cells_as_width
        else:
            self.__total_cells_as_width = self.__max_prefix_columns
            if self.__total_cells_as_width > 0:
                self.__total_cells_as_width += 1
            self.__total_cells_as_width += self.__max_score_columns
            if self.__max_columns > self.__total_cells_as_width:
                self.__total_cells_as_width = self.__max_columns
            self.__total_cells_as_width += 2
            print(f"width in cells: {self.__total_cells_as_width}")
            width = self.__total_cells_as_width * self.__cell_width

        self.__total_cells_as_height = (len(self.__parsed_lines) + 1 + 2)
        print(f"height in cells: {self.__total_cells_as_height}")

        height = self.__cell_height * self.__total_cells_as_height

        self.__paint = Paint(width, height, mode="RGB", background_color="white")
        print(f'width={width} height={height}')

    def __draw_lines(self):
        line_index = 1
        self.__left_offset = 0
        for line in self.__parsed_lines:
            if isinstance(line, ParsedLineAsTitle):
                self.__draw_title_line(line_index, line)
            elif isinstance(line, ParsedLineAsScore):
                self.__draw_score_line(line_index, line)
            elif isinstance(line, ParsedLineAsLyric):
                self.__draw_lyric_line(line_index, line)
            elif isinstance(line, ParsedLineAsBlank):
                pass
            else:
                self.__paint.text(
                    (self.__cell_width, line_index * self.__cell_height),
                    "不支持的 PARSED LINE",
                    font=self.__options.get_default_font(),
                    align="center"
                )

            line_index += 1

    def __draw_title_line(self, line_index: int, title_line: ParsedLineAsTitle):
        if title_line.get_left_component() is not None:
            # box = self.__paint.textbbox(
            #     xy=(0, 0),
            #     text=title_line.get_left_componet(),
            #     font=self.__options.get_font_for_title_left_component(),
            #     align="left",
            #     anchor="lm"
            # )
            self.__paint.text(
                xy=(self.__cell_width, int((line_index + 0.5) * self.__cell_height)),
                text=title_line.get_left_component(),
                font=self.__options.get_font_for_title_left_component(),
                align="left",
                fill="black",
                anchor="lm"
            )
        if title_line.get_middle_component() is not None:
            # box = self.__paint.textbbox(
            #     xy=(0, 0),
            #     text=title_line.get_middle_component(),
            #     font=self.__options.get_font_for_title_middle_component(),
            #     align="center",
            #     anchor="mm"
            # )
            self.__paint.text(
                xy=(
                    self.__paint.get_image().width / 2,  # - box[2] / 2,
                    int((line_index + 0.5) * self.__cell_height)
                ),
                text=title_line.get_middle_component(),
                font=self.__options.get_font_for_title_middle_component(),
                align="center",
                fill="black",
                anchor="mm"
            )
        if title_line.get_right_component() is not None:
            box = self.__paint.textbbox(
                xy=(0, 0),
                text=title_line.get_right_component(),
                font=self.__options.get_font_for_title_right_component(),
                align="right",
                anchor="lm"
            )
            xy = (
                self.__paint.get_image().width - self.__cell_width * 1 - box[2],
                int((line_index + 0.5) * self.__cell_height)
            )
            self.__paint.text(
                xy=xy,
                text=title_line.get_right_component(),
                font=self.__options.get_font_for_title_right_component(),
                align="right",
                fill="black",
                anchor="lm"
            )

    def __draw_score_line(self, line_index: int, score_line: ParsedLineAsScore):
        cell_index = 1
        if self.__max_prefix_columns > 0:
            cell_index += self.__max_prefix_columns + 1

        in_tie_or_slur = False
        ts_start_cell_index = None
        ts_end_cell_index = None
        ts_above_y = None

        this_score_columns = 0
        for su in score_line.get_score_units():
            this_score_columns += su.get_cell_needed()

        if self.__max_prefix_columns > 0:
            # 1 | PREFIX + 1 + X + REAL + X + 1 = WIDTH
            y = self.__total_cells_as_width - 2 - this_score_columns - self.__max_prefix_columns
            x = y / 2
            self.__left_offset = int((self.__total_cells_as_width - 1 - x - this_score_columns) * self.__cell_width)
            # self.__left_offset = int((self.__max_prefix_columns + (self.__max_score_columns - this_score_columns) / 2) * self.__cell_width)
        else:
            # 1 | X + REAL + X + 1 = WIDTH
            x = (self.__total_cells_as_width - 1 - this_score_columns) / 2
            self.__left_offset = int((self.__total_cells_as_width - 1 - x - this_score_columns) * self.__cell_width)
            # self.__left_offset = int(((self.__max_score_columns - this_score_columns + 1) / 2) * self.__cell_width)

        for i in range(len(score_line.get_score_units())):
            # debug cells outline
            # self.__paint.rectangle(
            #     xy=(
            #         (self.__left_offset+(cell_index) * self.__cell_width, line_index * self.__cell_height),
            #         (self.__left_offset+(cell_index + 1) * self.__cell_width, (line_index + 1) * self.__cell_height),
            #     ),
            #     outline="red"
            # )

            su = score_line.get_score_unit(i)
            # print(f"now at cell [{cell_index}] to draw su note [{su.get_note()}]")

            # draw su
            note = su.get_note()
            if note == "(":
                in_tie_or_slur = True
                ts_start_cell_index = cell_index
                ts_above_y = None
                continue
            elif note == ")":
                in_tie_or_slur = False
                # ts_end_cell_index = cell_index
                ts_end_cell_index += 1
                # draw tie or clur
                self.aimed_tie_or_slur(
                    (self.__left_offset + (ts_start_cell_index + 0.4) * self.__cell_width, ts_above_y),
                    (self.__left_offset + (ts_end_cell_index - 0.4) * self.__cell_width, ts_above_y),
                    int(max(5.0, self.__cell_height * 0.2)),
                    "black"
                )
                continue
            elif note == ")3":
                in_tie_or_slur = False
                # ts_end_cell_index = cell_index
                ts_end_cell_index += 1
                # draw tie or clur for triplet
                self.aimed_tie_or_slur(
                    (self.__left_offset + (ts_start_cell_index + 0.4) * self.__cell_width, ts_above_y),
                    (self.__left_offset + (ts_end_cell_index - 0.4) * self.__cell_width, ts_above_y),
                    int(max(5.0, self.__cell_height * 0.2)),
                    "black"
                )
                self.aimed_text(
                    int(self.__left_offset + (ts_start_cell_index + ts_end_cell_index) / 2 * self.__cell_width),
                    int(ts_above_y - max(5.0, self.__cell_height * 0.2) - 2),
                    "3",
                    self.__options.get_font_for_score_note(),
                    "black",
                    background_color="white"
                )
                continue
            elif note in ("|", "||", "||:", ":||", "|:", ":|"):
                self.__print_text_to_score_cell(
                    note,
                    cell_index,
                    line_index,
                    self.__left_offset
                )
                # self.aimed_text(
                #     int((cell_index + 0.5) * self.__cell_width),
                #     int((line_index + 0.5) * self.__cell_height),
                #     text=note,
                #     font=self.__font,
                #     fill="black"
                # )
                cell_index += 1
            else:
                # 0, 1-7
                self.__print_text_to_score_cell(
                    note,
                    cell_index,
                    line_index,
                    self.__left_offset
                )
                # self.aimed_text(
                #     int((cell_index + 0.5) * self.__cell_width),
                #     int((line_index + 0.5) * self.__cell_height),
                #     text=note,
                #     font=self.__font,
                #     fill="black"
                # )

                if in_tie_or_slur:
                    ts_end_cell_index = cell_index

                under_line_start_cell_index = cell_index
                if su.get_right_line_count() > 0:
                    for right_line_index in range(su.get_right_line_count()):
                        cell_index += 1
                        self.__paint.line(
                            xy=(
                                (
                                    int(self.__left_offset + (cell_index + 0.2) * self.__cell_width),
                                    int((line_index + 0.55) * self.__cell_height)
                                ),
                                (
                                    int(self.__left_offset + (cell_index + 0.8) * self.__cell_width),
                                    int((line_index + 0.55) * self.__cell_height)
                                )
                            ),
                            fill="black",
                            width=self.__line_width
                        )
                    # for right_line_index in range(su.get_right_line_count()):
                    #     cell_index += 1
                    #     self.__print_text_to_score_cell(
                    #         "－",
                    #         cell_index,
                    #         line_index,
                    #         self.__left_offset
                    #     )
                if su.get_dot():
                    # cell_index += 1
                    self.aimed_point(
                        self.__left_offset + int((cell_index + 0.9) * self.__cell_width),
                        int((line_index + 0.6) * self.__cell_height),
                        self.__point_width,
                        "black"
                    )
                under_line_end_cell_index = cell_index

                under_y = int((line_index + 1 - 0.2) * self.__cell_height)
                if su.get_under_line_count() > 0:
                    x0 = under_line_start_cell_index * self.__cell_width
                    x1 = (under_line_end_cell_index + 1) * self.__cell_width
                    for under_line_index in range(su.get_under_line_count()):
                        self.__paint.line(
                            xy=(
                                (self.__left_offset + x0, under_y), (self.__left_offset + x1, under_y)
                            ),
                            fill="black",
                            width=self.__line_width
                        )
                        under_y += int(1 + self.__line_width * 1.5)
                if su.get_below_point_count() > 0:
                    for below_point_index in range(su.get_below_point_count()):
                        self.aimed_point(
                            self.__left_offset + int((under_line_start_cell_index + 0.5) * self.__cell_width),
                            under_y,
                            self.__point_width,
                            "black"
                        )
                        under_y += int(1 + self.__point_width * 1.5)

                above_y = int((line_index + 0.25) * self.__cell_height)
                if su.get_above_point_count() > 0:
                    for above_point_index in range(su.get_above_point_count()):
                        self.aimed_point(
                            self.__left_offset + int((under_line_start_cell_index + 0.5) * self.__cell_width),
                            above_y,
                            self.__point_width,
                            "black"
                        )
                        above_y -= int(1 + self.__point_width * 1.5)

                if su.get_fermata():
                    self.aimed_fermata(
                        self.__left_offset + int((under_line_start_cell_index + 0.5) * self.__cell_width),
                        above_y,
                        "black"
                    )
                    above_y = above_y - self.__cell_width / 3 - self.__point_width

                if in_tie_or_slur:
                    if ts_above_y is None or ts_above_y > above_y:
                        ts_above_y = above_y

                if su.get_remark_text() is not None and su.get_remark_text() != "":
                    if su.get_remark_style() == ":<":
                        self.__paint.text(
                            xy=(
                                self.__left_offset + under_line_start_cell_index * self.__cell_width,
                                line_index * self.__cell_height
                            ),
                            text=su.get_remark_text(),
                            font=self.__options.get_font_for_score_note(),
                            align="left",
                            fill="black",
                            anchor="lm"
                        )
                    elif su.get_remark_style() == ":=":
                        self.__paint.text(
                            xy=(
                                int(self.__left_offset + (under_line_start_cell_index + 0.5) * self.__cell_width),
                                line_index * self.__cell_height
                            ),
                            text=su.get_remark_text(),
                            font=self.__options.get_font_for_score_note(),
                            align="left",
                            fill="black",
                            anchor="mm"
                        )
                    elif su.get_remark_style() == ":>":
                        self.__paint.text(
                            xy=(
                                self.__left_offset + (under_line_start_cell_index + 1) * self.__cell_width,
                                line_index * self.__cell_height
                            ),
                            text=su.get_remark_text(),
                            font=self.__options.get_font_for_score_note(),
                            align="left",
                            fill="black",
                            anchor="rm"
                        )

                cell_index += 1

    def __draw_lyric_line(self, line_index: int, lyric_line: ParsedLineAsLyric):
        lyric_body_offset = self.__cell_width
        if self.__max_prefix_columns > 0:
            if lyric_line.get_prefix() is not None:
                box = self.__paint.textbbox(
                    xy=(0, 0),
                    text=lyric_line.get_prefix(),
                    font=self.__options.get_font_for_lyric_prefix(),
                    align="center"
                )
                self.__paint.text(
                    xy=(
                        self.__cell_width * (1 + self.__max_prefix_columns) - box[2] + box[0],
                        line_index * self.__cell_height
                    ),
                    text=lyric_line.get_prefix(),
                    font=self.__options.get_font_for_lyric_prefix(),
                    align="right",
                    fill="black"
                )

            lyric_body_offset += (self.__max_prefix_columns + 1) * self.__cell_width

        left_offset = 0
        if lyric_line.get_content_align_type() == ">":
            left_offset = self.__left_offset
        elif lyric_line.get_content_align_type() == "=":
            left_offset = int(
                (self.__max_score_columns - len(lyric_line.get_content().strip())) / 2.0 * self.__cell_width)

        for i in range(len(lyric_line.get_content())):
            self.__paint.text(
                xy=(
                    left_offset + lyric_body_offset + i * self.__cell_width,
                    line_index * self.__cell_height
                ),
                text=lyric_line.get_lyric_char(i),
                font=self.__options.get_font_for_lyric(),
                align="center",
                fill="black"
            )

    def __print_text_to_score_cell(self, text, cell_index, line_index, left_offset: int = 0):
        self.aimed_text(
            int((cell_index + 0.5) * self.__cell_width + left_offset),
            int((line_index + 0.5) * self.__cell_height),
            text,
            self.__options.get_font_for_score()
        )

    def aimed_text(self, center_x, center_y, text, font, fill="black", background_color=None):
        box = self.__paint.textbbox(
            xy=(0, 0),
            text=text,
            font=font,
            align="center"
        )
        x = int(center_x - box[2] / 2)
        y = int(center_y - box[3] / 2)

        if background_color is not None:
            self.__paint.rectangle(
                ((x, y), (x + box[2], y + box[3])),
                fill=background_color
            )

        self.__paint.text(
            xy=(x, y),
            text=text,
            fill=fill,
            font=font,
            align="center"
        )

    def aimed_point(self, center_x, center_y, width: int = 3, fill="black"):
        self.__paint.ellipse(
            xy=(
                (int(center_x - width / 2), int(center_y - width / 2)),
                (int(center_x + width / 2), int(center_y + width / 2)),
            ),
            fill=fill,
            width=width
        )

    def aimed_fermata(self, point_x, point_y, fill="black"):
        self.__paint.arc(
            (
                (point_x - self.__cell_width / 3, point_y - self.__cell_width / 3),
                (point_x + self.__cell_width / 3, point_y)
            ),
            180,
            0,
            fill=fill,
            width=self.__point_width
        )
        self.aimed_point(point_x, int(point_y - self.__cell_width / 12), self.__point_width, fill)

    def aimed_tie_or_slur(self, start_xy, end_xy, high, fill="black"):
        self.__paint.arc(
            (
                (int(start_xy[0]), int(start_xy[1] - high)),
                (int(end_xy[0]), int(end_xy[1]))
            ),
            180, 0,
            fill=fill,
            width=self.__line_width
        )



import socket

import psutil


def get_local_ip():
    addresses = psutil.net_if_addrs()

    for interface, addr_list in addresses.items():
        for addr in addr_list:
            if addr.family == socket.AF_INET:
                ip_address = addr.address
                if ip_address != '127.0.0.1':
                    return ip_address

    return '127.0.0.1'


if __name__ == '__main__':
    print(get_local_ip())


import socket
from abc import abstractmethod
from threading import Thread
from typing import Optional

from nehushtan.socket.SocketHandleThreadManager import SocketHandlerThreadManager


class NehushtanUDPSocketServer:
    """
    Since 0.4.16
    """

    def __init__(self, host: str, port: int, buffer_size=0, tm: SocketHandlerThreadManager = None):
        self.__listen_port = port
        self.__listen_host = host
        self.__socket_instance: Optional[socket.socket] = None
        if buffer_size <= 0:
            buffer_size = 1024
        self.__buffer_size = buffer_size
        if not tm:
            tm = SocketHandlerThreadManager()
        self.__thread_manager = tm

    def get_thread_manager(self):
        return self.__thread_manager

    def get_socket_instance(self):
        return self.__socket_instance

    def listen(self):
        self.__socket_instance = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.__socket_instance.bind((self.__listen_host, self.__listen_port))

        while not self.should_terminate():
            data, address = self.__socket_instance.recvfrom(self.__buffer_size)

            thread = Thread(target=self.handle_incoming_data, name=address, args=(data, address))
            self.__thread_manager.register_thread(thread)
            thread.start()

    def should_terminate(self) -> bool:
        """
        If return True, stop the listener.
        Override it to implement it with your own logic.
        """
        return False

    @abstractmethod
    def handle_incoming_data(self, data: bytes, address):
        pass



from abc import abstractmethod

from nehushtan.logger.NehushtanLogger import NehushtanLogger


class NehushtanMPJob:
    """
    Since 0.2.13
    """

    def __init__(self, name: str):
        """
        Override this and extend the parameters
        Since 0.2.16, remove `logger`
        """
        self.__name = name
        self.__pid = 0

    def get_name(self):
        return self.__name

    @abstractmethod
    def get_logger(self) -> NehushtanLogger:
        """
        Since 0.2.16
        """
        pass

    def set_pid(self, pid: int):
        """
        Available in Main Process
        """
        self.__pid = pid
        return self

    def get_pid(self):
        """
        Available in Main Process
        """
        return self.__pid

    @abstractmethod
    def handle(self):
        """
        Run in sub process
        """
        pass

    def when_exited(self, exit_code: int):
        """
        Run in main process
        It is called when main process found worker process exited
        """
        pass



import math
from typing import List

from nehushtan.wave.Note import Note
from nehushtan.wave.WaveWriter import WaveWriter


class SimpleComposer:
    def __init__(self, target_wav_file_path: str, beats_in_minute: int = 96):
        self.__wave_writer = WaveWriter(target_wav_file_path)
        self.__beats_in_minute = beats_in_minute

        self.__notes_lines = {0: []}
        # self.__note_frames_cache = []

    def add_note_to_line(self, note: Note, line_index: int = 0):
        if not self.__notes_lines.get(line_index):
            self.__notes_lines[line_index] = []
        self.__notes_lines[line_index].append(note)

    def add_notes_to_lines(self, notes: List[Note]):
        for line_index in range(len(notes)):
            self.add_note_to_line(notes[line_index], line_index)

    def __process_notes(self):
        # 每次采样代表的秒数 = 1秒 / 每秒采样次数
        step = 1.0 / self.__wave_writer.get_frame_rate()
        mixed_frames = []
        print("TO PROCESS NOTES")
        time_piece_index = 0
        for line_index, line in self.__notes_lines.items():
            line_frames = []
            for note in line:
                hz = note.get_sound_in_hz()
                beats = note.get_beats()
                # 本音占用时长 = 拍数 * (一分钟 / 每分钟拍数)
                time = beats * 60.0 / self.__beats_in_minute
                piece_count = int(round(time / step))

                for i in range(piece_count):
                    t = step * time_piece_index

                    frame = math.cos(2 * math.pi * (hz * t))
                    frame = frame * 10000  # it is like volumn

                    line_frames.append(int(frame))
                    time_piece_index += 1

            for i in range(len(line_frames)):
                frame = line_frames[i]
                if line_index == 0:
                    mixed_frames.append(frame)
                else:
                    mixed_frames[i] += frame

        print("PREPARED")

        self.__wave_writer.write_frames(mixed_frames)

    def close(self):
        self.__process_notes()
        self.__wave_writer.close()
        print("WRITTEN AND CLOSED")



import socket
from abc import abstractmethod
from typing import Optional


class NehushtanTCPSocketClient:
    """
    Since 0.4.16
    """

    def __init__(self, port: int, host=None):
        self.__listen_port = port
        self.__listen_host = host
        self.__socket_instance: Optional[socket.socket] = None

    def listen(self):
        address_entries = socket.getaddrinfo(
            self.__listen_host,
            self.__listen_port,
            socket.AF_UNSPEC,
            socket.SOCK_STREAM
        )

        prepare_error = None

        for address_entry in address_entries:
            af, socket_type, proto, canon_name, sa = address_entry
            try:
                self.__socket_instance = socket.socket(af, socket_type, proto)
                self.__socket_instance.connect(sa)
            except OSError as msg:
                prepare_error = msg
                if self.__socket_instance is not None:
                    self.__socket_instance.close()
                self.__socket_instance = None
                continue
            break

        if self.__socket_instance is None:
            raise prepare_error

        with self.__socket_instance:
            self.handle_client_connection()

    def get_connection(self):
        return self.__socket_instance

    @abstractmethod
    def handle_client_connection(self):
        """
        Override it to implement it with your own logic.
        """
        pass

    def close_connection(self):
        if self.__socket_instance is not None:
            self.__socket_instance.close()
        self.__socket_instance = None



import select
import sys
import time

if __name__ == '__main__':
    print(f"now: {time.time()}")
    readable, _, _ = select.select([sys.stdin], [], [], 5)
    print(f"now: {time.time()}")
    print(readable)

    print(f"now: {time.time()}")
    readable, _, _ = select.select([], [], [], 2)
    print(f"now: {time.time()}")
    print(readable)



import logging
from datetime import datetime

from nehushtan.logger.legacy.NehushtanFileLogger import NehushtanFileLogger
from nehushtan.logger.legacy.NehushtanLogging import NehushtanLogging
from nehushtan.logger.NehushtanLogger import NehushtanLogger, NehushtanLoggerAdapterWithStdOut, \
    NehushtanLoggerAdapterWithFileWriter
from tests.config import LOGGER_CONFIG

log_dir = LOGGER_CONFIG['log_dir']


def test_nehushtan_file_logger():
    file_logger = NehushtanFileLogger('test', log_dir, log_level=logging.INFO,
                                      print_higher_than_this_level=NehushtanLogging.INFO,
                                      record_millisecond=True)

    file_logger.critical('C', {"A": "B"})
    file_logger.error('E', 111)
    file_logger.warning('W', 'dd')
    file_logger.info('I', file_logger)
    file_logger.debug('D')

    try:
        raise ValueError('A<>B')
    except Exception as error:
        file_logger.exception('X', error)

    args_json = file_logger.get_args_json_to_clone()
    print(args_json)
    cloned_logger = NehushtanFileLogger.build_instance_from_args_json(args_json)

    cloned_logger.critical('C', {"A": "B"})
    cloned_logger.error('E', 111)
    cloned_logger.warning('W', 'dd')
    cloned_logger.info('I', cloned_logger)
    cloned_logger.debug('D')

    try:
        raise ValueError('A<>B')
    except Exception as error:
        cloned_logger.exception('X', error)

    progress_logger = NehushtanFileLogger()
    total = 36
    for i in range(total):
        progress_logger.log_progress('TEST', i, total, desc=f'now done[{i}]')


def test_nehushtan_logger_with_stdout_adapter():
    logger = NehushtanLogger(
        topic='default',
        adapter=NehushtanLoggerAdapterWithStdOut(),
    )
    logger.info('kakaka', {"d": "g"})


def test_nehushtan_logger_with_file_writer_adapter():
    logger = NehushtanLogger(
        topic='test/file/log',
        adapter=NehushtanLoggerAdapterWithFileWriter(
            log_dir=log_dir,
            record_millisecond=True
        )
    )
    logger.info('kakaka', {"d": "g"})

    try:
        d = {"k": datetime(year=2024, month=1, day=1), }
        logger.notice('date', d)
        p = d['p']
    except Exception as error:
        logger.exception('error', error)


if __name__ == '__main__':
    test_nehushtan_logger_with_stdout_adapter()
    test_nehushtan_logger_with_file_writer_adapter()



import time
from datetime import datetime

from aliyun.log import LogClient, LogItem, PutLogsRequest

from nehushtan.helper.CommonHelper import CommonHelper
from nehushtan.logger.AliyunSLSLoggerAdapter import AliyunSLSLoggerAdapter
from nehushtan.logger.NehushtanLogger import NehushtanLogger
from tests.config import ALIYUN_SLS_CONFIG

endpoint = ALIYUN_SLS_CONFIG.get('endpoint')  # 选择与上面步骤创建Project所属区域匹配的Endpoint
accessKeyId = ALIYUN_SLS_CONFIG.get('accessKeyId')  # 使用你的阿里云访问密钥AccessKeyId
accessKey = ALIYUN_SLS_CONFIG.get('accessKeySecret')  # 使用你的阿里云访问密钥AccessKeySecret
project = ALIYUN_SLS_CONFIG.get('project')  # 上面步骤创建的项目名称
logstore = ALIYUN_SLS_CONFIG.get('logstore')  # 上面步骤创建的日志库名称

topic = "NehushtanTest"
source: str = ALIYUN_SLS_CONFIG.get('source')


# source = source.replace('[IP]', CommonHelper.get_local_ip())


def test_raw():
    # 构建一个client
    client = LogClient(endpoint, accessKeyId, accessKey)
    print(f'client: {client}')

    logItem = LogItem()
    logItem.set_time(int(time.time()))
    logItem.set_contents([
        ("message", CommonHelper.generate_random_uuid_hex()),
    ])

    req2 = PutLogsRequest(
        project=project,
        logstore=logstore,
        topic=topic,
        source=source.replace('[IP]', CommonHelper.get_local_ip()),
        logitems=[logItem]
    )
    res2 = client.put_logs(req2)
    res2.log_print()


def test_class():
    logger = NehushtanLogger(
        topic=topic,
        adapter=AliyunSLSLoggerAdapter(
            endpoint=endpoint,
            access_key_id=accessKeyId,
            access_key_secret=accessKey,
            project=project,
            logstore=logstore,
            source=source,
        )
    )

    logger.info('long live sinri', {'now': datetime.now()})

    try:
        d = {}
        p = d['p']
    except Exception as e:
        logger.exception('miao!', e)


if __name__ == '__main__':
    test_class()